#! /usr/bin/env python

"""
This class is designed to provide functions to allow the automated
identification of planar facet surfaces above fault traces.
Module is SLOW (e.g., minutes+ per full analysis of a "large" data set). It is
only intended for model post-analysis or DEM analysis. Do not loop this class!!
This is part of the NSF funded project investigating fault scarp degradation,
Tucker, Hobley, McCoy.
"""
from __future__ import print_function
import numpy as np
import sys
from pylab import plot, colorbar, figure, show
from scipy.stats import mode
from six.moves import range

from landlab.plot import imshow as gridshow
from landlab.utils import structured_grid as sgrid


class find_facets(object):
    """
    Note that this class assumes the grid does not change during the model
    run. Changes to data stored in the grid should (?) update automatically.

    If *fault_azimuth* is supplied, it should be -pi/2 < az <= pi/2 (i.e.,
    we don't consider fault dip, even if it's known).
    """

    def __init__(self, grid, elev_field='topographic__elevation',
                 fault_azimuth=None):
        """
        Note that this class assumes the grid does not change during the model
        run. Changes to data stored in the grid should (?) update automatically.

        If *fault_azimuth* is supplied, it should be -pi/2 < az <= pi/2 (i.e.,
        we don't consider fault dip, even if it's known).
        """
        if not np.isclose(grid.dx, grid.dy):
            raise ValueError('row and column spacing must be the same')

        self.grid = grid
        self.elevs = self.grid.at_node[elev_field]
        self.az = fault_azimuth

    def analyse_fault_trace(self, fault_trace_node_ids):
        """
        This method takes the grid and an array listing the (contiguous) node
        ids of cells that contain a single fault segment trace of interest.

        It sets and returns the azimuth of the fault trace, az,
        -pi/2 < az <= pi/2.
        (i.e., smallest angle between north and the trace).
        """
        self.ft_trace_node_ids = fault_trace_node_ids
        x = self.grid.node_x[fault_trace_node_ids]
        y = self.grid.node_y[fault_trace_node_ids]
        (grad, offset) = np.polyfit(x, y, 1)
        angle = np.arctan(grad)
        if grad >= 0.:
            self.az = np.pi / 2. - angle
        else:
            self.az = -np.pi / 2. - angle

        return self.az

    def set_slopes_aspects(self):
        self.slopes = self.grid.calc_slopes_of_nodes(elevs=self.elevs)
        self.aspect = self.grid.calc_aspect_of_node(elevs=self.elevs)
        print('Calculated and stored slopes and aspects...')

    def define_aspect_node_subset(self, angle_tolerance=5.):
        """
        This method sets and returns a list of all nodes in the landscape which
        have an
        aspect within 5 degrees of perpendicular to the fault trace.

        It assumes self.az, the angle between north and the fault trace, has
        already been set, and also that self.slopes and self.aspect are also
        set.
        The returned boolean array is num_core_nodes long.
        *angle_tolerance* is the angle in degrees that the aspect must be within
        from the fault trace angle.
        NB: this version is too discriminating on the aspect restriction,
        presumably because we use only a single ft angle for what's really
        a 2d trace. Need to work with local aspect.
        """
        perp_to_az = np.pi / 2. + self.az
        five_degrees = np.pi / 180. * angle_tolerance  # note the silly naming
        perp_plus_five = (perp_to_az + five_degrees) % (2. * np.pi)
        perp_minus_five = (perp_to_az - five_degrees) % (2. * np.pi)
        other_dip_plus_five = (perp_to_az + np.pi +
                               five_degrees) % (2. * np.pi)
        other_dip_minus_five = (perp_to_az + np.pi -
                                five_degrees) % (2. * np.pi)

        # need to be careful near the 2pi->0 discontinuity...
        greater_condition = np.greater(self.aspect, perp_minus_five)
        lesser_condition = np.less(self.aspect, perp_plus_five)
        if (perp_to_az - five_degrees) < 0.:
            condition_first_dip = np.logical_or(
                greater_condition, lesser_condition)
        else:
            condition_first_dip = np.logical_and(
                greater_condition, lesser_condition)
        greater_condition_2 = np.greater(self.aspect, other_dip_minus_five)
        lesser_condition_2 = np.less(self.aspect, other_dip_plus_five)
        if (perp_to_az + np.pi + five_degrees) // (2. * np.pi):  # top condition exceeds 2pi
            condition_opposite_dip = np.logical_or(
                greater_condition_2, lesser_condition_2)
        else:
            condition_opposite_dip = np.logical_and(
                greater_condition_2, lesser_condition_2)

        self.aspect_close_nodes = np.logical_or(
            condition_first_dip, condition_opposite_dip)
        print('Calculated and stored nodes with aspects compatible with fault trace...')
        return self.aspect_close_nodes

    def define_aspect_node_subset_local(self, dist_tolerance=4., angle_tolerance=15., dip_dir='E'):
        """
        """
        grid = self.grid
        try:
            print('using subset')
            # remember, steep_nodes is already core_nodes.size long
            subset = np.where(self.steep_nodes)[0]
        except NameError:
            print('using all nodes')
            subset = np.arange(grid.core_nodes.size)
        closest_ft_node = np.empty(subset.size, dtype=int)
        angle_to_ft = np.empty_like(closest_ft_node, dtype=float)
        new_angle_to_ft = np.empty_like(closest_ft_node, dtype=float)
        distance_to_ft = np.empty_like(closest_ft_node, dtype=float)
        distance_to_ft.fill(sys.float_info.max)
        new_distance_to_ft = np.empty_like(closest_ft_node, dtype=float)
        for i in self.ft_trace_node_ids:
            grid.calc_distances_of_nodes_to_point((grid.node_x[i], grid.node_y[i]),
                                                 node_subset=grid.core_nodes[
                                                     subset], get_az='angles',
                                                 out_distance=new_distance_to_ft, out_azimuth=new_angle_to_ft)
            closer_nodes = new_distance_to_ft < distance_to_ft
            distance_to_ft[closer_nodes] = new_distance_to_ft[closer_nodes]
            angle_to_ft[closer_nodes] = new_angle_to_ft[closer_nodes]
            closest_ft_node[closer_nodes] = i
        self.closest_ft_node = -np.ones(grid.core_nodes.size)
        self.distance_to_ft = -np.ones(grid.core_nodes.size)
        self.angle_to_ft = -np.ones(grid.core_nodes.size)
        self.closest_ft_node[subset] = closest_ft_node
        self.distance_to_ft[subset] = distance_to_ft
        # angle_to_ft is actually the angle_from_ft! So we need to adjust.
        # a second problem is that pts downslope (opposite az) can also be on the line.
        # solution - take a dip_dir input...
        angle_to_ft = (angle_to_ft + np.pi) % (2. * np.pi)
        self.angle_to_ft[subset] = angle_to_ft
        #gridshow.imshow_grid_at_node(self.grid, self.distance_to_ft)
        # show()
        #gridshow.imshow_grid_at_node(self.grid, self.angle_to_ft)
        # show()
        # the relevant condition is now that the local aspect and angle to fault
        # are the same...
        # We need to bias the five degrees against distant points, as it's easier
        # to have similar angles in the far field. Rule should be in px - the
        # two angles should be within *angle_tol* px of each other at the ft
        # trace.
        divergence_at_ft = distance_to_ft * \
            np.tan((angle_to_ft - self.aspect[subset]) % np.pi)
        # might be *too* forgiving for close-in nodes
        condition = np.less(np.fabs(divergence_at_ft),
                            grid.dx * dist_tolerance)
        #...so add another tester; must be w/i 15 degrees of each other:
        diff_angles = np.min([np.fabs(angle_to_ft - self.aspect[subset]), np.fabs(
            np.fabs(angle_to_ft - self.aspect[subset]) - 2. * np.pi)], axis=0)
        self.diff_angles = np.empty(grid.core_nodes.size, dtype=float)
        self.diff_angles.fill(sys.float_info.max)
        self.diff_angles[subset] = diff_angles
        #gridshow.imshow_grid_at_node(self.grid, self.angle_to_ft)
        # show()
        figure(6)
        gridshow.imshow_grid_at_node(self.grid, np.where(
            self.diff_angles < 100000., self.diff_angles, -1.))
        condition2 = np.less(diff_angles, angle_tolerance * np.pi / 180.)
        condition = np.logical_and(condition, condition2)
        core_nodes_size_condition = np.zeros(grid.core_nodes.size, dtype=bool)
        core_nodes_size_condition[subset] = condition
        #gridshow.imshow_grid_at_node(self.grid, core_nodes_size_condition)
        # show()
        #core_nodes_size_condition = np.zeros(grid.core_nodes.size, dtype=bool)
        #core_nodes_size_condition[subset] = condition2
        #gridshow.imshow_grid_at_node(self.grid, core_nodes_size_condition)
        # show()
        self.aspect_close_nodes = core_nodes_size_condition
        print('Calculated and stored nodes with aspects compatible with fault trace...')
        return self.aspect_close_nodes

    def define_steep_nodes(self, threshold_in_degrees=5.):
        """
        This method sets and returns a list of all nodes in the landscape which
        are "steep" and could be part of a facet.
        The critical hillslope angle is set by *threshold_in_degrees*, and
        defaults to 5.

        This assumes you have already called define_aspect_node_subset, in
        which self.slope is set.
        The returned boolean array is num_core_nodes long.
        """
        threshold_in_rads = threshold_in_degrees * np.pi / 180.
        self.steep_nodes = np.greater(self.slopes, threshold_in_rads)
        print('Calculated and stored nodes with slopes exceeding slope threshold...')
        #gridshow.imshow_grid_at_node(self.grid, self.steep_nodes)
        # show()
        return self.steep_nodes

    def show_possible_nodes(self):
        """
        Once the subsets by aspect and slope have been set, call this function
        to see both the whole elevation map, and the subset of nodes that
        will be searched.
        """
        possible_core_nodes = np.logical_and(
            self.steep_nodes, self.aspect_close_nodes)
        figure(1)
        gridshow.imshow_grid_at_node(self.grid, self.elevs)
        figure(2)
        gridshow.imshow_grid_at_node(self.grid, self.slopes)
        figure(3)
        gridshow.imshow_grid_at_node(self.grid, self.aspect)
        figure(4)
        gridshow.imshow_grid_at_node(self.grid, possible_core_nodes)
        show()

    def find_coherent_facet_patches(self, tolerance=3., threshold_num_px=12):
        """
        This method searches the (already determined) possible pixels for
        patches with coherent slope angles, within a *tolerance* (in degrees).
        A patch is only recorded if it consists of at least *threshold_num_px*.

        The method records and returns:

        *  a ragged array of lists, where each list is the pixels comprising
           each facet patch, and
        *  a (num_patches, 2) array recording the mean slope and and its stdev
           for each patch.
        """
        self.possible_core_nodes = np.where(np.logical_and(
            self.steep_nodes, self.aspect_close_nodes))[0]
        consistent_slope_patches = []
        for i in self.possible_core_nodes:
            nodes_in_patch = np.array([i])
            patch_size = 1
            mean_slope = self.slopes[nodes_in_patch]
            while 1:
                possible_neighbors = np.union1d(
                    self.grid.active_neighbors_at_node(
                        nodes_in_patch).flat, self.possible_core_nodes)
                neighbor_slopes = self.slopes[possible_neighbors]
                low_tol_condition = np.greater(
                    neighbor_slopes, mean_slope - tolerance)
                high_tol_condition = np.less(
                    neighbor_slopes, mean_slope + tolerance)
                total_condition = np.logical_and(
                    low_tol_condition, high_tol_condition)
                nodes_in_patch = possible_neighbors[total_condition]
                new_patch_size = nodes_in_patch.size
                if patch_size == new_patch_size:
                    break
                else:
                    patch_size = new_patch_size
                    mean_slope = np.mean(neighbor_slopes[total_condition])
            if new_patch_size > threshold_num_px:
                consistent_slope_patches.append(nodes_in_patch)
        # May also need a uniqueness test: a px should only appear in one list.
        #(i.e., all patches containing a given px should all be identical)
        self.consistent_slope_patches = consistent_slope_patches
        return consistent_slope_patches

    def find_slope_lines(self, tolerance=1.):
        """
        This method attempts to find slope-consistent line profiles up facets,
        perpendicular to the fault.
        Assumes you used define_aspect_node_subset_local().
        """
        grid = self.grid
        self.possible_core_nodes = np.where(np.logical_and(
            self.steep_nodes, self.aspect_close_nodes))[0]
        pcn = self.possible_core_nodes
        unique_starting_pts = np.unique(
            self.closest_ft_node[pcn])  # in real node nos
        # set up places to store the profile data:
        profile_ft_node_id = []
        profile_ft_node_x = []
        profile_ft_node_y = []
        profile_ft_node_z = []
        profile_ft_node_dist = []
        profile_x_facet_pts = []
        profile_z_facet_pts = []
        profile_S_facet_pts = []
        count = 0
        for i in unique_starting_pts:
            count += 1
            print("Running ", count, " of ", unique_starting_pts.size)
            # set the local angle of the ft trace:
            ft_pt_distances_to_node = self.grid.calc_distances_of_nodes_to_point((grid.node_x[i], grid.node_y[i]),
                                                                                node_subset=self.ft_trace_node_ids)
            close_ft_nodes = np.less(ft_pt_distances_to_node, 5. * grid.dx)
            x = grid.node_x[self.ft_trace_node_ids[close_ft_nodes]]
            y = grid.node_y[self.ft_trace_node_ids[close_ft_nodes]]
            (grad, offset) = np.polyfit(x, y, 1)
            condition = np.equal(self.closest_ft_node[pcn], i)
            nodes_possible = pcn[condition]
            print(nodes_possible.size, " nodes")
            if nodes_possible.size > 10.:
                #their_az = self.angle_to_ft[nodes_possible]
                #their_diff_angles = self.diff_angles[nodes_possible]
                their_elevs = self.elevs[grid.core_nodes][nodes_possible]
                #their_distances = self.distance_to_ft[nodes_possible]
                # need to make new distances so we remove the ambiguity of angle around the ft point (i.e., dists from a far-field pt on the ft normal)
                # now make a multiplier to make sure the reference point for
                # new distances is far from the actual pts:
                multiplier = 10. * \
                    np.ptp(grid.node_y[grid.core_nodes[nodes_possible]])
                # derive the position:
                x_ref = grid.node_x[i] + cmp(grid.node_x[i], np.mean(
                    grid.node_x[grid.core_nodes[nodes_possible]])) * multiplier * abs(grad)
                y_ref = grid.node_y[
                    i] + cmp(grid.node_y[i], np.mean(grid.node_y[grid.core_nodes[nodes_possible]])) * multiplier
                # get new absolute distances
                dist_to_ft = self.grid.calc_distances_of_nodes_to_point(
                    (x_ref, y_ref), node_subset=np.array([i]))
                dists_along_profile = self.grid.calc_distances_of_nodes_to_point(
                    (x_ref, y_ref), node_subset=grid.core_nodes[nodes_possible]) - dist_to_ft
                # note the ft is now the origin, but pts might be back-to-front (consistently, though)
                # sort the distances. Remove any pts that aren't in a "cluster".
                # We assume there will be one big "bunched" plane, then a load
                # of outliers
                dist_order = np.argsort(dists_along_profile)
                dist_diffs = np.diff(dists_along_profile[dist_order])
                print("dists along profile sorted: ",
                      dists_along_profile[dist_order])
                print("dist diffs: ", dist_diffs)
                # max_diff = 3.*np.median(dist_diffs) #######this might need
                # attention if there's a heavy tail on the distances
                if grad < 1:
                    mod = np.sqrt(1. + grad**2.)
                else:
                    mod = np.sqrt(1. + (1. / grad)**2.)
                max_diff = 1.9 * mod * grid.dx
                locs_of_large_diffs = np.where(dist_diffs > max_diff)[0]
                # there should only be 1 place on the line where there's a cluster, i.e., a large pts_betw_of_max_diffs.
                # This is what we're seeking.
                # ...this can be empty quite easily
                pts_betw_large_diffs = np.diff(locs_of_large_diffs)
                # need to be careful here in case the where call gives an empty
                # array
                if locs_of_large_diffs.size > 1:
                    biggest_interval_loc = np.argmax(pts_betw_large_diffs)
                elif locs_of_large_diffs.size == 1:
                    # one side or the other must be bigger:
                    if 2. * locs_of_large_diffs[0] < dists_along_profile.size - 1:
                        locs_of_large_diffs = np.array(
                            [locs_of_large_diffs[0], (dists_along_profile.size - 1)])
                    else:
                        locs_of_large_diffs = np.array(
                            [0, locs_of_large_diffs[0]])
                    biggest_interval_loc = np.array([0])
                    # here we assume that the single large diff must be further
                    # from the ft than the plane
                else:
                    locs_of_large_diffs = np.array(
                        [0, (dists_along_profile.size - 1)])
                    biggest_interval_loc = np.array([0])
                    #...all the pts in the line are one cluster
                # apply a test to ensure we only save "big" patches; a
                # threshold of 10 pts on the line
                try:
                    patch_size = pts_betw_large_diffs[biggest_interval_loc]
                except IndexError:  # pts_betw_large_diffs is empty
                    patch_size = locs_of_large_diffs[
                        1] - locs_of_large_diffs[0]
                if patch_size > 10.:
                    start_pt_of_cluster = locs_of_large_diffs[
                        biggest_interval_loc] + 1
                    end_pt_of_cluster = locs_of_large_diffs[
                        biggest_interval_loc + 1] + 1  # both referring to the sorted list
                    # both +1s are to account for required frame of ref changes - indices refer to where the big gaps start, not where they ends
                    # so:
                    dists_to_sorted_pts = dists_along_profile[
                        dist_order][start_pt_of_cluster:end_pt_of_cluster]
                    elevs_of_sorted_pts = their_elevs[dist_order][
                        start_pt_of_cluster:end_pt_of_cluster]
                    slopes_of_sorted_pts = self.slopes[nodes_possible][
                        dist_order][start_pt_of_cluster:end_pt_of_cluster]
                    profile_ft_node_id.append(i.copy())
                    profile_ft_node_x.append(grid.node_x[i].copy())
                    profile_ft_node_y.append(grid.node_y[i].copy())
                    profile_ft_node_z.append(self.elevs[i].copy())
                    profile_ft_node_dist.append(dist_to_ft.copy())
                    profile_x_facet_pts.append(dists_to_sorted_pts.copy())
                    profile_z_facet_pts.append(elevs_of_sorted_pts.copy())
                    profile_S_facet_pts.append(slopes_of_sorted_pts.copy())
                    figure(5)
                    plot(dists_to_sorted_pts, elevs_of_sorted_pts)
                    # dirty, but effective code!

        self.profile_ft_node_id = profile_ft_node_id
        self.profile_ft_node_x = profile_ft_node_x
        self.profile_ft_node_y = profile_ft_node_y
        self.profile_ft_node_z = profile_ft_node_z
        self.profile_ft_node_dist = profile_ft_node_dist
        self.profile_x_facet_pts = profile_x_facet_pts
        self.profile_z_facet_pts = profile_z_facet_pts
        self.profile_S_facet_pts = profile_S_facet_pts

    def fit_slopes_to_facet_lines(self, polynomial_degree=4, curvature_threshold=0.0004):
        """
        Fits (linear) lines of best fit to extracted profiles, already stored as
        class properties.
        """
        avg_slopes_linear = []
        avg_slopes_poly = []
        curv_of_flattest_part_list = []
        slope_min_curv = []
        rsqd_list = []
        big_slope_small_curv = []
        elev_at_bssc = []
        for i in range(len(self.profile_x_facet_pts)):
            x = self.profile_x_facet_pts[i]
            z = self.profile_z_facet_pts[i]
            (grad, offset) = np.polyfit(x, z, 1)
            coeffs, residuals = np.polyfit(
                x, z, polynomial_degree, full=True)[:2]
            rsqd = 1. - residuals / (z.size * z.var())
            # differentiate the coeffs to get slope:
            diff_multiplier = np.arange(polynomial_degree + 1)[::-1]
            curv_multiplier = np.arange(polynomial_degree)[::-1]
            z_equ = np.poly1d(coeffs)
            S_equ = np.poly1d((coeffs * diff_multiplier)[:-1])
            curv_equ = np.poly1d(
                ((coeffs * diff_multiplier)[:-1] * curv_multiplier)[:-1])
            S_at_each_pt = S_equ(x)
            curv_at_each_pt = curv_equ(x)
            avg_slopes_linear.append(abs(grad))
            avg_slopes_poly.append(np.amax(np.fabs(S_at_each_pt)))
            loc_of_flattest_part = np.argmin(
                np.fabs(curv_at_each_pt[2:-2])) + 2
            curv_of_flattest_part = curv_at_each_pt[loc_of_flattest_part]
            S_at_min_curve_untested = abs(S_at_each_pt[loc_of_flattest_part])
            small_curves = np.less(
                np.fabs(curv_at_each_pt[2:-2]), curvature_threshold)
            try:
                big_slope_small_curv.append(
                    np.amax(S_at_each_pt[small_curves]))
                elev_at_bssc.append(z[np.argmax(S_at_each_pt[small_curves])])
            except ValueError:
                big_slope_small_curv.append(np.nan)
                elev_at_bssc.append(np.nan)
            slope_min_curv.append(S_at_min_curve_untested)
            curv_of_flattest_part_list.append(curv_of_flattest_part)
            rsqd_list.append(rsqd)
            # figure(8)
            #synthetic_z = grad*x + offset
            synthetic_z = z_equ(x)
            plot(x, z, 'x')
            plot(x, synthetic_z, '-')
        self.avg_slopes_linear = np.array(avg_slopes_linear)
        self.avg_slopes_poly = np.array(avg_slopes_poly)
        self.curv_of_flattest_part = np.array(curv_of_flattest_part_list)
        self.slope_min_curv = np.array(slope_min_curv)
        self.big_slope_small_curv = np.array(big_slope_small_curv)
        self.elev_at_bssc = np.array(elev_at_bssc)
        self.rsqd = np.array(rsqd_list)



import numpy as np

from . import links
from ..base import BAD_INDEX_VALUE


def number_of_faces(shape):
    """Number of faces in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of faces in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.faces import number_of_faces
    >>> number_of_faces((3, 4))
    7
    """
    if len(shape) != 2:
        raise ValueError('shape must be size 2')

    if min(shape) > 2:
        return ((shape[0] - 1) * (shape[1] - 2) +
                (shape[0] - 2) * (shape[1] - 1))
    else:
        return 0


def face_at_link(shape, bad_index_value=BAD_INDEX_VALUE):
    """Get face for each link of a grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Examples
    --------
    >>> from landlab.grid.structured_quad.faces import face_at_link
    >>> face_at_link((4, 5), bad_index_value=-1)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([-1, -1, -1, -1, -1,  0,  1,  2, -1,
            3,  4,  5,  6, -1,  7,  8,  9, -1,
           10, 11, 12, 13, -1, 14, 15, 16, -1,
           -1, -1, -1, -1])
    """
    from .c_faces import _face_at_link

    out = np.full(links.number_of_links(shape), bad_index_value, dtype=int)
    _face_at_link(shape, out)

    return out


def link_at_face(shape):
    """Get the link at each face.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Examples
    --------
    >>> from landlab.grid.structured_quad.faces import link_at_face
    >>> link_at_face((3, 5)) # doctest: +NORMALIZE_WHITESPACE
    array([  5,  6,  7,
             9, 10, 11, 12,
            14, 15, 16])

    Notes
    -----
    The algorithm is based on the following considerations:

    1. To get the link ID associated with a face ID, we start with the face
       ID and add something to it.
    2. Face 0 crosses the second vertical link. The first NC-1 links are
       horizontal, and run along the bottom of the grid. Then there's a
       vertical link on the lower-left side of the grid. Thus the link that
       crosses face 0 will be ID number NC, where NC is the number of cols.
    3. The offset between face ID and link ID increases by 1 every time we
       finish a row of horizontal faces (NC-2 of them) with vertical links,
       and again when we finish a row of vertical faces (NC-1 of them) with
       horizontal links. Together, a "row" of horizontal and vertical faces
       makes up 2NC-3 faces; this is the variable "fpr" (for "faces per 
       row") below.
    4. The quantity 2 * (face_ids // fpr) increases by 2 for every "full"
       (horizontal plus vertical) row of faces.
    5. The quantity ((face_ids % fpr) >= (NC-2)) increases by 1 every time
       we shift from a horizontal to a vertical row of faces (because
       each "full row" has NC-2 horizontal faces, then NC-1 vertical ones.)
    6. So, to find the offset, we add the "basic" offset, NC, to the "full
       row" offset, 2*(face_ids//fpr), and the "mid-row" offset, 
       ((face_ids % fpr)>=(NC-2)).
    """
    faces = np.arange(number_of_faces(shape))

    faces_per_row = (2 * shape[1]) - 3
    links = (shape[1] + (2 * (faces // faces_per_row)) +
             ((faces % faces_per_row) >= (shape[1] - 2)) + faces)

    return links



import numpy as np
from numpy.testing import assert_array_equal
from nose import with_setup
try:
    from nose.tools import assert_is
except ImportError:
    from landlab.testing.tools import assert_is

from landlab import RasterModelGrid


_GRIDS = {}


def setup_grids():
    """Set up test grids with unit and non-unit spacing."""
    _GRIDS.update({
        'unit': RasterModelGrid(4, 5),
        'non_unit': RasterModelGrid(4, 5, 2.),
        'non_square': RasterModelGrid((4, 5), spacing=(5, 2)),
    })


@with_setup(setup_grids)
def test_unit_spacing():
    """Test with a grid with unit spacing."""
    rmg, values_at_nodes = _GRIDS['unit'], np.arange(20)
    grads = rmg.calc_grad_at_link(values_at_nodes)[rmg.active_links]

    assert_array_equal(grads,
                       np.array([5.0, 5.0, 5.0,
                                 1.0, 1.0, 1.0, 1.0,
                                 5.0, 5.0, 5.0,
                                 1.0, 1.0, 1.0, 1.0,
                                 5.0, 5.0, 5.0,]))

    diffs = rmg.calculate_diff_at_active_links(values_at_nodes)
    assert_array_equal(grads, diffs)


@with_setup(setup_grids)
def test_non_unit_spacing():
    """Test with a grid with non-unit spacing."""
    rmg, values_at_nodes = _GRIDS['non_square'], np.arange(20)

    grads = rmg.calc_grad_at_link(values_at_nodes)[rmg.active_links]
    assert_array_equal(grads,
                       np.array([1.0, 1.0, 1.0,
                                 0.5, 0.5, 0.5, 0.5,
                                 1.0, 1.0, 1.0,
                                 0.5, 0.5, 0.5, 0.5,
                                 1.0, 1.0, 1.0]))
    diffs = rmg.calculate_diff_at_active_links(values_at_nodes)
    assert_array_equal(diffs,
                       np.array([5.0, 5.0, 5.0,
                                 1.0, 1.0, 1.0, 1.0,
                                 5.0, 5.0, 5.0,
                                 1.0, 1.0, 1.0, 1.0,
                                 5.0, 5.0, 5.0,]))


@with_setup(setup_grids)
def test_out_array():
    """Test using the out keyword."""
    rmg, values_at_nodes = _GRIDS['non_square'], np.arange(20)

    output_array = np.empty(rmg.number_of_links)
    rtn_array = rmg.calc_grad_at_link(values_at_nodes, out=output_array)
    assert_array_equal(rtn_array[rmg.active_links],
                       np.array([1.0, 1.0, 1.0,
                                 0.5, 0.5, 0.5, 0.5,
                                 1.0, 1.0, 1.0,
                                 0.5, 0.5, 0.5, 0.5,
                                 1.0, 1.0, 1.0]))
    assert_is(rtn_array, output_array)


@with_setup(setup_grids)
def test_diff_out_array():
    """Test returned array is the same as that passed as out keyword."""
    rmg = RasterModelGrid(4, 5)
    values = np.arange(20)
    diff = np.empty(17)
    rtn_diff = rmg.calculate_diff_at_active_links(values, out=diff)
    assert_array_equal(
        diff,
        np.array([5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5]))
    assert_is(rtn_diff, diff)



def set_status_at_node_on_edges(grid, right=None, top=None, left=None,
                                bottom=None):
    """Set node status on grid edges.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    right : int, optional
        Node status along right edge.
    top : int, optional
        Node status along top edge.
    left : int, optional
        Node status along left edge.
    bottom : int, optional
        Node status along bottom edge.

    Examples
    --------
    >>> from landlab import RasterModelGrid, CLOSED_BOUNDARY

    >>> grid = RasterModelGrid((3, 4))
    >>> grid.status_at_node # doctest: +NORMALIZE_WHITESPACE
    array([1, 1, 1, 1,
           1, 0, 0, 1,
           1, 1, 1, 1], dtype=int8)

    >>> grid.set_status_at_node_on_edges(right=CLOSED_BOUNDARY)
    >>> grid.status_at_node # doctest: +NORMALIZE_WHITESPACE
    array([1, 1, 1, 4,
           1, 0, 0, 4,
           1, 1, 1, 4], dtype=int8)

    >>> from landlab import FIXED_GRADIENT_BOUNDARY
    >>> grid = RasterModelGrid((3, 4))

    The status of a corner is set along with its clockwise edge. That is,
    if setting the status for the top and right edges, the upper-right corner
    has the status of the right edge.

    >>> grid.set_status_at_node_on_edges(top=CLOSED_BOUNDARY,
    ...     right=FIXED_GRADIENT_BOUNDARY)
    >>> grid.status_at_node # doctest: +NORMALIZE_WHITESPACE
    array([1, 1, 1, 2,
           1, 0, 0, 2,
           4, 4, 4, 2], dtype=int8)

    In the above example, if you wanted the corner to have the status of the
    top edge, you need to make two calls to `set_status_at_node_on_edges`,

    >>> grid = RasterModelGrid((3, 4))
    >>> grid.set_status_at_node_on_edges(right=FIXED_GRADIENT_BOUNDARY)
    >>> grid.set_status_at_node_on_edges(top=CLOSED_BOUNDARY)
    >>> grid.status_at_node # doctest: +NORMALIZE_WHITESPACE
    array([1, 1, 1, 2,
           1, 0, 0, 2,
           4, 4, 4, 4], dtype=int8)

    An example that sets all of the edges shows how corners are set.

    >>> grid.set_status_at_node_on_edges(right=1, top=2, left=3, bottom=4)
    >>> grid.status_at_node # doctest: +NORMALIZE_WHITESPACE
    array([3, 4, 4, 4,
           3, 0, 0, 1,
           2, 2, 2, 1], dtype=int8)
    """
    status_at_edge = (('bottom', bottom), ('left', left), ('top', top),
                      ('right', right), )

    for edge, val in status_at_edge:
        if val is not None:
            nodes = grid.nodes_at_edge(edge)
            grid.status_at_node[nodes] = val

    if right is not None and bottom is not None:
        lr = grid.nodes_at_right_edge[0]
        grid.status_at_node[lr] = bottom



#! /usr/bin/env python
"""Read/write data from an ESRI ASCII file into a RasterModelGrid.

ESRI ASCII functions
++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.io.esri_ascii.read_asc_header
    ~landlab.io.esri_ascii.read_esri_ascii
    ~landlab.io.esri_ascii.write_esri_ascii
"""

import os
import re
import six

import numpy as np

_VALID_HEADER_KEYS = [
    'ncols', 'nrows', 'xllcorner', 'xllcenter', 'yllcorner',
    'yllcenter', 'cellsize', 'nodata_value',
]
_HEADER_KEY_REGEX_PATTERN = re.compile(r'\s*(?P<key>[a-zA-z]\w+)')
_HEADER_REGEX_PATTERN = re.compile(
    r'\s*(?P<key>[a-zA-Z]\w+)\s+(?P<value>[\w.+-]+)')
_HEADER_VALUE_TESTS = {
    'nrows': (int, lambda x: x > 0),
    'ncols': (int, lambda x: x > 0),
    'cellsize': (float, lambda x: x > 0),
    'xllcorner': (float, lambda x: True),
    'xllcenter': (float, lambda x: True),
    'yllcorner': (float, lambda x: True),
    'yllcenter': (float, lambda x: True),
    'nodata_value': (float, lambda x: True),
}


class Error(Exception):

    """Base class for errors in this module."""

    pass


class BadHeaderLineError(Error):

    """Raise this error for a bad header is line."""

    def __init__(self, line):
        self._line = line

    def __str__(self):
        return self._line


class MissingRequiredKeyError(Error):

    """Raise this error when a header is missing a required key."""

    def __init__(self, key):
        self._key = key

    def __str__(self):
        return self._key


class KeyTypeError(Error):

    """Raise this error when a header's key value is of the wrong type."""

    def __init__(self, key, expected_type):
        self._key = key
        self._type = str(expected_type)

    def __str__(self):
        return 'Unable to convert %s to %s' % (self._key, self._type)


class KeyValueError(Error):

    """Raise this error when a header's key value has a bad value."""

    def __init__(self, key, message):
        self._key = key
        self._msg = message

    def __str__(self):
        return '%s: %s' % (self._key, self._msg)


class DataSizeError(Error):

    """Raise this error if the size of data does not match the header."""

    def __init__(self, size, expected_size):
        self._actual = size
        self._expected = expected_size

    def __str__(self):
        return '%s != %s' % (self._actual, self._expected)


class MismatchGridDataSizeError(Error):
    
    """Raise this error if the data size does not match the grid size."""
    
    def __init__(self, size, expected_size):
        self._actual = size
        self._expected = expected_size

    def __str__(self):
        return '(data size) %s != %s (grid size)' \
           % (self._actual, self._expected)

    
def _parse_header_key_value(line):
    """Parse a header line into a key-value pair.

    Parameters
    ----------
    line : str
        Header line.

    Returns
    -------
    (str, str)
        Header key-value pair

    Raises
    ------
    BadHeaderLineError
        The is something wrong with the header line.
    """
    match = _HEADER_KEY_REGEX_PATTERN.match(line)
    if match is None:
        return None
        # raise BadHeaderLineError(line)

    match = _HEADER_REGEX_PATTERN.match(line)
    if match is None:
        raise BadHeaderLineError(line)

    (key, value) = (match.group('key').lower(), match.group('value'))

    if key in _VALID_HEADER_KEYS:
        return (key, value)
    else:
        raise BadHeaderLineError(line)


def _header_lines(asc_file):
    """Iterate over header lines for a ESRI ASCII file.

    Parameters
    ----------
    asc_file : file_like
        File-like object for an ESRI ASCII file.

    Yields
    ------
    str
        Header line.
    """
    pos = asc_file.tell()
    line = asc_file.readline()
    while len(line) > 0:
        if len(line.strip()) > 0:
            item = _parse_header_key_value(line)
            if item:
                yield item
            else:
                asc_file.seek(pos, 0)
                break
        pos = asc_file.tell()
        line = asc_file.readline()


def _header_is_valid(header):
    """Check if the ESRI ASCII header is valid.

    Parameters
    ----------
    header : dict
        Header as key-values pairs.

    Raises
    ------
    MissingRequiredKeyError
        The header is missing a required key.
    KeyTypeError
        The header has the key but its values is of the wrong type.
    """
    header_keys = set(header)
    required_keys = set(['ncols', 'nrows', 'cellsize'])

    if not required_keys.issubset(header_keys):
        raise MissingRequiredKeyError(', '.join(required_keys - header_keys))

    for keys in [('xllcenter', 'xllcorner'), ('yllcenter', 'yllcorner')]:
        if len(set(keys) & header_keys) != 1:
            raise MissingRequiredKeyError('|'.join(keys))

    for (key, requires) in _HEADER_VALUE_TESTS.items():
        to_type, is_valid = requires

        if key not in header:
            continue

        try:
            header[key] = to_type(header[key])
        except ValueError:
            raise KeyTypeError(key, to_type)

        if not is_valid(header[key]):
            raise KeyValueError(key, 'Bad value')

    return True


def read_asc_header(asc_file):
    """Read header information from an ESRI ASCII raster file.

    The header contains the following variables,
        - *ncols*: Number of cell columns
        - *nrows*: Number of cell rows
        - *xllcenter* or *xllcorner*: X (column) coordinate of lower-left
            coordinate of grid (by center or lower-left corner of the cell)
        - *yllcenter*, *yllcorner*: Y (row) coordinate of lower-left
            coordinate of grid (by center or lower-left corner of the cell)
        - *cellsize*: Grid spacing between rows and columns
        - *nodata_value*: No-data value (optional)

    Parameters
    ----------
    asc_file : file_like
        File-like object from which to read header.

    Returns
    -------
    dict
        Header as key-value pairs.

    Raises
    ------
    MissingRequiredKeyError
        The header is missing a required key.
    KeyTypeError
        The header has the key but its values is of the wrong type.

    Examples
    --------
    >>> from six import StringIO
    >>> from landlab.io.esri_ascii import read_asc_header
    >>> contents = StringIO('''
    ...     nrows 100
    ...     ncols 200
    ...     cellsize 1.5
    ...     xllcenter 0.5
    ...     yllcenter -0.5
    ... ''')
    >>> hdr = read_asc_header(contents)
    >>> hdr['nrows'], hdr['ncols']
    (100, 200)
    >>> hdr['cellsize']
    1.5
    >>> hdr['xllcenter'], hdr['yllcenter']
    (0.5, -0.5)

    ``MissingRequiredKey`` is raised if the header does not contain all of the
    necessary keys.

    >>> contents = StringIO('''
    ...     ncols 200
    ...     cellsize 1.5
    ...     xllcenter 0.5
    ...     yllcenter -0.5
    ... ''')
    >>> read_asc_header(contents) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    MissingRequiredKeyError: nrows

    ``KeyTypeError`` is raises if a value is of the wrong type. For instance,
    ``nrows`` and ``ncols`` must be ``int``.

    >>> contents = StringIO('''
    ...     nrows 100.5
    ...     ncols 200
    ...     cellsize 1.5
    ...     xllcenter 0.5
    ...     yllcenter -0.5
    ... ''')
    >>> read_asc_header(contents) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    KeyTypeError: Unable to convert nrows to <type 'int'>
    """
    header = dict()
    for (key, value) in _header_lines(asc_file):
        header[key] = value

    _header_is_valid(header)

    return header


def _read_asc_data(asc_file):
    """Read gridded data from an ESRI ASCII data file.

    Parameters
    ----------
    asc_file : file-like
        File-like object of the data file pointing to the start of the data.

    .. note::
        First row of the data is at the top of the raster grid, the second
        row is the second from the top, and so on.
    """
    return np.loadtxt(asc_file)


def read_esri_ascii(asc_file, grid=None, reshape=False, name=None, halo=0):
    """Read :py:class:`~landlab.RasterModelGrid` from an ESRI ASCII file.

    Read data from *asc_file*, an ESRI_ ASCII file, into a
    :py:class:`~landlab.RasterModelGrid`.  *asc_file* is either the name of
    the data file or is a file-like object.

    The grid and data read from the file are returned as a tuple
    (*grid*, *data*) where *grid* is an instance of
    :py:class:`~landlab.RasterModelGrid` and *data* is a numpy
    array of doubles with that has been reshaped to have the number of rows
    and columns given in the header.

    .. _ESRI: http://resources.esri.com/help/9.3/arcgisengine/java/GP_ToolRef/spatial_analyst_tools/esri_ascii_raster_format.htm

    Parameters
    ----------
    asc_file : str of file-like
        Data file to read.
    reshape : boolean, optional
        Reshape the returned array, otherwise return a flattened array.
    name : str, optional
        Add data to the grid as a named field.
    grid : *grid* , optional
        Adds data to an existing *grid* instead of creating a new one.
    halo : integer, optional
        Adds outer border of depth halo to the *grid*. 

    Returns
    -------
    (grid, data) : tuple
        A newly-created RasterModel grid and the associated node data.
        
    Raises
    ------
    DataSizeError
        Data are not the same size as indicated by the header file.
    MismatchGridDataSizeError
        If a grid is passed, the size of the grid does not agree with the
        size of the data.
        
    Examples
    --------
    Assume that fop is the name of a file that contains text below
    (make sure you have your path correct):
    ncols         3
    nrows         4
    xllcorner     1.
    yllcorner     2.
    cellsize      10.
    NODATA_value  -9999
    0. 1. 2.
    3. 4. 5.
    6. 7. 8.
    9. 10. 11.
    --------
    >>> from landlab.io import read_esri_ascii
    >>> (grid, data) = read_esri_ascii('fop') # doctest: +SKIP
    >>> #grid is an object of type RasterModelGrid with 4 rows and 3 cols
    >>> #data contains an array of length 4*3 that is equal to
    >>> # [9., 10., 11., 6., 7., 8., 3., 4., 5., 0., 1., 2.]
    >>> (grid, data) = read_esri_ascii('fop', halo=1) # doctest: +SKIP
    >>> #now the data has a nodata_value ring of -9999 around it. So array is
    >>> # [-9999, -9999, -9999, -9999, -9999, -9999,
    >>> #  -9999, 9., 10., 11., -9999, 
    >>> #  -9999, 6., 7., 8., -9999, 
    >>> #  -9999, 3., 4., 5., -9999,
    >>> #  -9999, 0., 1., 2. -9999,
    >>> #  -9999, -9999, -9999, -9999, -9999, -9999]
    """
    from ..grid import RasterModelGrid

    if isinstance(asc_file, six.string_types):
        file_name = asc_file
        with open(file_name, 'r') as asc_file:
            header = read_asc_header(asc_file)
            data = _read_asc_data(asc_file)
    else:
        header = read_asc_header(asc_file)
        data = _read_asc_data(asc_file)
    
    #There is no reason for halo to be negative.
    #Assume that if a negative value is given it should be 0.
    if halo <= 0:
        shape = (header['nrows'], header['ncols'])
        if data.size != shape[0] * shape[1]:
            raise DataSizeError(shape[0] * shape[1], data.size)
    else:
        shape = (header['nrows'] + 2 * halo, header['ncols'] + 2 * halo)
        #check to see if a nodata_value was given.  If not, assign -9999.
        if 'nodata_value' in header.keys():
            nodata_value = header['nodata_value']
        else:
            header['nodata_value'] = -9999.
            nodata_value = header['nodata_value']
        if data.size != (shape[0] - 2 * halo) * (shape[1] - 2 * halo):
            raise DataSizeError(shape[0] * shape[1], data.size)
    spacing = (header['cellsize'], header['cellsize'])
    #origin = (header['xllcorner'], header['yllcorner'])   
    
    data = np.flipud(data)

    #REMEMBER, shape contains the size with halo in place
    #header contains the shape of the original data
    #Add halo below
    if halo > 0:
        helper_row = np.ones(shape[1]) * nodata_value
        #for the first halo row(s), add num cols worth of nodata vals to data
        for i in range(0, halo):
            data = np.insert(data,0,helper_row)
        #then for header['nrows'] add halo number nodata vals, header['ncols'] 
        #of data, then halo number of nodata vals
        helper_row_ends = np.ones(halo) * nodata_value
        for i in range(halo, header['nrows']+halo):
            #this adds at the beginning of the row
            data = np.insert(data,i * shape[1],helper_row_ends)
            #this adds at the end of the row
            data = np.insert(data,(i + 1) * shape[1] - halo,helper_row_ends)
        #for the last halo row(s), add num cols worth of nodata vals to data
        for i in range(header['nrows']+halo,shape[0]):
            data = np.insert(data,data.size,helper_row)
        
    if not reshape:
        data = data.flatten()
        
    if grid is not None:
        if (grid.number_of_node_rows != shape[0]) or \
        (grid.number_of_node_columns != shape[1]):
            raise MismatchGridDataSizeError(shape[0] * shape[1], \
            grid.number_of_node_rows * grid.number_of_node_columns )

    if grid is None:
        grid = RasterModelGrid(shape, spacing=spacing)
    if name:
        grid.add_field('node', name, data)

    return (grid, data)


def write_esri_ascii(path, fields, names=None, clobber=False):
    """Write landlab fields to ESRI ASCII.

    Write the data and grid information for *fields* to *path* in the ESRI
    ASCII format.

    Parameters
    ----------
    path : str
        Path to output file.
    fields : field-like
        Landlab field object that holds a grid and associated values.
    names : iterable of str, optional
        Names of the fields to include in the output file. If not provided,
        write all fields.
    clobber : boolean
        If *path* exists, clobber the existing file, otherwise raise an
        exception.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.testing.tools import cdtemp
    >>> from landlab import RasterModelGrid
    >>> from landlab.io.esri_ascii import write_esri_ascii

    >>> grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    >>> _ = grid.add_field('node', 'air__temperature', np.arange(20.))
    >>> with cdtemp() as _:
    ...     files = write_esri_ascii('test.asc', grid)
    >>> files
    ['test.asc']

    >>> _ = grid.add_field('node', 'land_surface__elevation', np.arange(20.))
    >>> with cdtemp() as _:
    ...     files = write_esri_ascii('test.asc', grid)
    >>> files.sort()
    >>> files
    ['test_air__temperature.asc', 'test_land_surface__elevation.asc']
    """
    if os.path.exists(path) and not clobber:
        raise ValueError('file exists')

    if isinstance(names, six.string_types):
        names = [names]

    names = names or fields.at_node.keys()

    if len(names) == 1:
        paths = [path]
    elif len(names) > 1:
        (base, ext) = os.path.splitext(path)
        paths = [base + '_' + name + ext for name in names]
    else:
        raise ValueError('no node fields to write')

    bad_names = set(names) - set(fields.at_node.keys())
    if len(bad_names) > 0:
        raise ValueError('unknown field name(s): %s' % ','.join(bad_names))

    header = {
        'ncols': fields.number_of_node_columns,
        'nrows': fields.number_of_node_rows,
        'xllcorner': fields.node_x[0],
        'yllcorner': fields.node_y[0],
        'cellsize': fields.dx,
    }

    for path, name in zip(paths, names):
        header_lines = ['%s %s' % (key, str(val))
                        for key, val in list(header.items())]
        data = fields.at_node[name].reshape(header['nrows'], header['ncols'])
        np.savetxt(path, np.flipud(data), header=os.linesep.join(header_lines),
                   comments='')

    return paths



"""
Unit tests for landlab.components.pet.potential_evapotranspiration_field
"""
from nose.tools import assert_equal, assert_true, assert_raises, with_setup
from numpy.testing import assert_array_almost_equal
try:
    from nose.tools import assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is_instance
import numpy as np

from landlab import RasterModelGrid
from landlab.components.pet.potential_evapotranspiration_field \
             import PotentialEvapotranspiration


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def setup_grid():
    from landlab import RasterModelGrid
    grid = RasterModelGrid((20, 20), spacing=10e0)
    PET = PotentialEvapotranspiration(grid)
    globals().update({
        'PET': PotentialEvapotranspiration(grid)
    })


@with_setup(setup_grid)
def test_name():
    assert_equal(PET.name, 'Potential Evapotranspiration')


@with_setup(setup_grid)
def test_input_var_names():
    assert_equal(PET.input_var_names,
                 ('radiation__ratio_to_flat_surface',))


@with_setup(setup_grid)
def test_output_var_names():
    assert_equal(sorted(PET.output_var_names),
                 ['radiation__incoming_shortwave',
                  'radiation__net',
                  'radiation__net_longwave',
                  'radiation__net_shortwave',
                  'surface__potential_evapotranspiration_rate'])


@with_setup(setup_grid)
def test_var_units():
    assert_equal(set(PET.input_var_names) |
                 set(PET.output_var_names),
                 set(dict(PET.units).keys()))

    assert_equal(PET.var_units('radiation__incoming_shortwave'), 'W/m^2')
    assert_equal(PET.var_units('radiation__net'), 'W/m^2')
    assert_equal(PET.var_units('radiation__net_longwave'), 'W/m^2')
    assert_equal(PET.var_units('radiation__net_shortwave'), 'W/m^2')
    assert_equal(PET.var_units('radiation__ratio_to_flat_surface'), 'None')
    assert_equal(PET.var_units('surface__potential_evapotranspiration_rate'),
                 'mm')


@with_setup(setup_grid)
def test_grid_shape():
    assert_equal(PET.grid.number_of_node_rows, _SHAPE[0])
    assert_equal(PET.grid.number_of_node_columns, _SHAPE[1])


@with_setup(setup_grid)
def test_grid_x_extent():
    assert_equal(PET.grid.extent[1], (_SHAPE[1] - 1) * _SPACING[1])


@with_setup(setup_grid)
def test_grid_y_extent():
    assert_equal(PET.grid.extent[0], (_SHAPE[0] - 1) * _SPACING[0])


@with_setup(setup_grid)
def test_field_getters():
    for name in PET.grid['node']:
        field = PET.grid['node'][name]
        assert_is_instance(field, np.ndarray)
        assert_equal(field.shape,
                     (PET.grid.number_of_node_rows *
                      PET.grid.number_of_node_columns, ))
                      
    for name in PET.grid['cell']:
        field = PET.grid['cell'][name]
        assert_is_instance(field, np.ndarray)
        assert_equal(field.shape,
                     (PET.grid.number_of_cell_rows *
                      PET.grid.number_of_cell_columns, ))

    assert_raises(KeyError, lambda: PET.grid['not_a_var_name'])


@with_setup(setup_grid)
def test_field_initialized_to_zero():
    for name in PET.grid['node']:
        field = PET.grid['node'][name]
        assert_array_almost_equal(field, np.zeros(PET.grid.number_of_nodes))
    for name in PET.grid['cell']:
        field = PET.grid['cell'][name]
        assert_array_almost_equal(field, np.zeros(PET.grid.number_of_cells))


"""Landlab component that simulates relative incidence shortwave radiation
on sloped surface.

Landlab component that computes 1D and 2D total incident shortwave
radiation. This code also computes relative incidence shortwave radiation
compared to a flat surface. Ref: Flores-Cervantes et al., 2012

.. codeauthor:: Sai Nudurupati & Erkan Istanbulluoglu

Examples
--------
>>> import numpy as np
>>> from landlab import RasterModelGrid
>>> from landlab.components import Radiation

Create a grid on which to calculate incident shortwave radiation

>>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))

The grid will need some input data. To check the names of the fields
that provide the input to this component, use the *input_var_names*
class property.

>>> Radiation.input_var_names
('topographic__elevation',)

Check the units for the fields.

>>> Radiation.var_units('topographic__elevation')
'm'

Create the input fields.

>>> grid['node']['topographic__elevation'] = np.array([
...       0., 0., 0., 0.,
...       1., 1., 1., 1.,
...       2., 2., 2., 2.,
...       3., 4., 4., 3.,
...       4., 4., 4., 4.])

If you are not sure about one of the input or output variables, you can
get help for specific variables.

>>> Radiation.var_help('topographic__elevation')
name: topographic__elevation
description:
  elevation of the ground surface relative to some datum
units: m
at: node
intent: in

Check the output variable names

>>> sorted(Radiation.output_var_names) # doctest: +NORMALIZE_WHITESPACE
['radiation__incoming_shortwave',
 'radiation__net_shortwave',
 'radiation__ratio_to_flat_surface']

Instantiate the 'Radiation' component to work on this grid, and run it.

>>> rad = Radiation(grid)

Run the *update* method to update output variables with current time
>>> current_time = 0.5
>>> rad.update(current_time)

>>> rad.grid.at_cell['radiation__ratio_to_flat_surface']
array([ 0.38488566,  0.38488566,  0.33309785,  0.33309785,  0.37381705,
        0.37381705])

>>> rad.grid.at_cell['radiation__incoming_shortwave']
array([ 398.33664988,  398.33664988,  344.73895668,  344.73895668,
        386.88120966,  386.88120966])
"""

from landlab import Component
from ...utils.decorators import use_file_name_or_kwds
import numpy as np


_VALID_METHODS = set(['Grid'])


def _assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class Radiation(Component):

    """Compute 1D and 2D total incident shortwave radiation.

    Landlab component that computes 1D and 2D total incident shortwave
    radiation. This code also computes relative incidence shortwave radiation
    compared to a flat surface.

    Construction::

        Radiation(grid, method='Grid', cloudiness=0.2, latitude=34., 
                  albedo=0.2, solar_constant=1366.67, 
                  clearsky_turbidity=2., opt_airmass=0.)

    Parameters
    ----------
    grid: RasterModelGrid
        A grid.
    method: {'Grid'}, optional
        Currently, only default is available.
    cloudiness: float, optional
        Cloudiness.
    latitude: float, optional
        Latitude (radians).
    albedo: float, optional
        Albedo.
    solar_constant: float, optional
        Solar Constant (W/m^2).
    clearsky_turbidity: float, optional
        Clear sky turbidity.
    opt_airmass: float, optional
        Optical air mass.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import Radiation
    >>> import numpy as np

    >>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))
    >>> rad = Radiation(grid)
    >>> rad.name
    'Radiation'
    >>> rad.input_var_names
    ('topographic__elevation',)
    >>> sorted(rad.output_var_names) # doctest: +NORMALIZE_WHITESPACE
    ['radiation__incoming_shortwave',
     'radiation__net_shortwave',
     'radiation__ratio_to_flat_surface']
    >>> sorted(rad.units) # doctest: +NORMALIZE_WHITESPACE
    [('radiation__incoming_shortwave', 'W/m^2'),
     ('radiation__net_shortwave', 'W/m^2'),
     ('radiation__ratio_to_flat_surface', 'None'),
     ('topographic__elevation', 'm')]

    >>> rad.grid.number_of_node_rows
    5
    >>> rad.grid.number_of_node_columns
    4
    >>> rad.grid is grid
    True
    >>> np.all(grid.at_cell['radiation__ratio_to_flat_surface'] == 0.)
    True
    >>> np.all(grid.at_node['topographic__elevation'] == 0.)
    True

    >>> grid['node']['topographic__elevation'] = np.array([
    ...       0., 0., 0., 0.,
    ...       1., 1., 1., 1.,
    ...       2., 2., 2., 2.,
    ...       3., 4., 4., 3.,
    ...       4., 4., 4., 4.])
    >>> current_time = 0.5
    >>> rad.update(current_time)
    >>> np.all(grid.at_cell['radiation__ratio_to_flat_surface'] == 0.)
    False
    """

    _name = 'Radiation'

    _input_var_names = (
        'topographic__elevation',
    )

    _output_var_names = (
        'radiation__incoming_shortwave',
        'radiation__ratio_to_flat_surface',
        'radiation__net_shortwave',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'radiation__incoming_shortwave': 'W/m^2',
        'radiation__ratio_to_flat_surface': 'None',
        'radiation__net_shortwave': 'W/m^2',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'radiation__incoming_shortwave': 'cell',
        'radiation__ratio_to_flat_surface': 'cell',
        'radiation__net_shortwave': 'cell',
    }

    _var_doc = {
        'topographic__elevation':
            'elevation of the ground surface relative to some datum',
        'radiation__incoming_shortwave':
            'total incident shortwave radiation over the time step',
        'radiation__ratio_to_flat_surface':
            'ratio of total incident shortwave radiation on sloped surface \
             to flat surface',
        'radiation__net_shortwave':
            'net incident shortwave radiation over the time step',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, method='Grid', cloudiness=0.2,
                 latitude=34., albedo=0.2, solar_constant=1366.67,
                 clearsky_turbidity=2., opt_airmass=0., **kwds):
        """
        Parameters
        ----------
        grid : RasterModelGrid
            A grid.
        method : {'Grid'}, optional
            Currently, only default is available.
        cloudiness: float, optional
            Cloudiness.
        latitude: float, optional
            Latitude (Radians).
        albedo: float, optional
            Albedo.
        solar_constant: float, optional
            Solar Constant (W/m^2).
        clearsky_turbidity: float, optional
            Clear sky turbidity.
        opt_airmass: float, optional
            Optical air mass.
        """

        self._method = method
        self._N = cloudiness
        self._latitude = latitude
        self._A = albedo
        self._Io = solar_constant
        self._n = clearsky_turbidity
        self._m = opt_airmass

        _assert_method_is_valid(self._method)

        super(Radiation, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros(name, at='node', units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros(name, at='cell', units=self._var_units[name])

        if 'Slope' not in self.grid.at_cell:
            self.grid.add_zeros('Slope', at='cell', units='radians')

        if 'Aspect' not in self.grid.at_cell:
            self.grid.add_zeros('Aspect', at='cell', units='radians')

        self._nodal_values = self.grid['node']
        self._cell_values = self.grid['cell']
        self._slope, self._aspect = \
            grid.calculate_slope_aspect_at_nodes_burrough(
                vals='topographic__elevation')
#        self._slope = grid.calc_slope_of_node( \
#                                elevs = 'topographic__elevation')
#        self._aspect =
        self._cell_values['Slope'] = self._slope
        self._cell_values['Aspect'] = self._aspect

    def update(self, current_time, hour=12., **kwds):
        """Update fields with current loading conditions.

        Parameters
        ----------
        current_time: float
              Current time (years).
        hour: float, optional
              Hour of the day.
        """
        self._t = hour
        self._radf = self._cell_values['radiation__ratio_to_flat_surface']
        self._Rs = self._cell_values['radiation__incoming_shortwave']
        self._Rnet = self._cell_values['radiation__net_shortwave']

        self._julian = np.floor(
            (current_time - np.floor(current_time)) * 365.25) # Julian day

        self._phi = np.radians(self._latitude) # Latitude in Radians

        self._delta = 23.45 * np.radians(
            np.cos(2*np.pi / 365 * (172 - self._julian))) # Declination angle

        self._tau = (self._t + 12.0) * np.pi / 12.0 # Hour angle

        self._alpha = np.arcsin(np.sin(self._delta) * np.sin(self._phi) +
                                np.cos(self._delta) * np.cos(self._phi) *
                                np.cos(self._tau)) # Solar Altitude

        if self._alpha <= 0.25 * np.pi / 180.0: # If altitude is -ve,
            self._alpha = 0.25 * np.pi / 180.0 # sun is beyond the horizon

        self._Rgl = (self._Io * np.exp((-1) * self._n * (
            0.128 - 0.054 * np.log10(1. / np.sin(self._alpha)))*(
                1. / np.sin(self._alpha))))
        # Counting for Albedo, Cloudiness and Atmospheric turbidity

        self._phisun = (np.arctan(- np.sin(self._tau) / (np.tan(self._delta) *
                        np.cos(self._phi) - np.sin(self._phi) *
                        np.cos(self._tau)))) # Sun's Azhimuth

        if (self._phisun >= 0 and - np.sin(self._tau) <= 0):
            self._phisun = self._phisun + np.pi

        elif (self._phisun <= 0 and - np.sin(self._tau) >= 0):
            self._phisun = self._phisun + np.pi

        self._flat = (np.cos(np.arctan(0)) * np.sin(self._alpha) +
                      np.sin(np.arctan(0)) * np.cos(self._alpha) *
                      np.cos(self._phisun - 0)) # flat surface reference

        self._Rsflat = self._Rgl * self._flat
        # flat surface total incoming shortwave radiation

        self._Rnetflat = ((1 - self._A) * (1 - 0.65 * (self._N ** 2)) *
                          self._Rsflat)
        # flat surface Net incoming shortwave radiation

        self._sloped = (np.cos(self._slope) * np.sin(self._alpha) +
                        np.sin(self._slope) * np.cos(self._alpha) *
                        np.cos(self._phisun - self._aspect))

        self._radf = self._sloped / self._flat

        self._radf[self._radf <= 0.] = 0.
        self._radf[self._radf > 6.] = 6.

        self._Rs = self._Rsflat * self._radf
        # Sloped surface Toatl Incoming Shortwave Radn
        self._Rnet = self._Rnetflat * self._radf

        self._cell_values['radiation__ratio_to_flat_surface'] = self._radf
        self._cell_values['radiation__incoming_shortwave'] = self._Rs
        self._cell_values['radiation__net_shortwave'] = self._Rnet



#! /usr/env/python
"""
Python implementation of ModelGrid, a base class used to create and manage
grids for 2D numerical models.

Getting Information about a Grid
--------------------------------
The following attributes, properties, and methods provide data about the grid,
its geometry, and the connectivity among the various elements. Each grid
element has an ID number, which is also its position in an array that
contains information about that type of element. For example, the *x*
coordinate of node 5 would be found at `grid.node_x[5]`.

The naming of grid-element arrays is *attribute*`_at_`*element*, where
*attribute* is the name of the data in question, and *element* is the element
to which the attribute applies. For example, the property `node_at_cell`
contains the ID of the node associated with each cell. For example,
`node_at_cell[3]` contains the *node ID* of the node associated with cell 3.
The *attribute* is singular if there is only one value per element; for
example, there is only one node associated with each cell. It is plural when
there are multiple values per element; for example, the `faces_at_cell` array
contains multiple faces for each cell. Exceptions to these general rules are
functions that return indices of a subset of all elements of a particular type.
For example, you can obtain an array with IDs of only the core nodes using
`core_nodes`, while `active_links` provides an array of IDs of active links
(only). Finally, attributes that represent a measurement of something, such as
the length of a link or the surface area of a cell, are described using `_of_`,
as in the example `area_of_cell`.

Information about the grid as a whole
+++++++++++++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.axis_name
    ~landlab.grid.hex.HexModelGrid.axis_units
    ~landlab.grid.hex.HexModelGrid.from_dict
    ~landlab.grid.hex.HexModelGrid.hexplot
    ~landlab.grid.hex.HexModelGrid.move_origin
    ~landlab.grid.hex.HexModelGrid.node_axis_coordinates
    ~landlab.grid.hex.HexModelGrid.number_of_active_faces
    ~landlab.grid.hex.HexModelGrid.number_of_active_links
    ~landlab.grid.hex.HexModelGrid.number_of_cells
    ~landlab.grid.hex.HexModelGrid.number_of_core_cells
    ~landlab.grid.hex.HexModelGrid.number_of_core_nodes
    ~landlab.grid.hex.HexModelGrid.number_of_elements
    ~landlab.grid.hex.HexModelGrid.number_of_faces
    ~landlab.grid.hex.HexModelGrid.number_of_fixed_links
    ~landlab.grid.hex.HexModelGrid.number_of_links
    ~landlab.grid.hex.HexModelGrid.number_of_node_columns
    ~landlab.grid.hex.HexModelGrid.number_of_node_rows
    ~landlab.grid.hex.HexModelGrid.number_of_nodes
    ~landlab.grid.hex.HexModelGrid.number_of_patches
    ~landlab.grid.hex.HexModelGrid.save

Information about nodes
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.active_link_dirs_at_node
    ~landlab.grid.hex.HexModelGrid.all_node_azimuths_map
    ~landlab.grid.hex.HexModelGrid.all_node_distances_map
    ~landlab.grid.hex.HexModelGrid.boundary_nodes
    ~landlab.grid.hex.HexModelGrid.cell_area_at_node
    ~landlab.grid.hex.HexModelGrid.cell_at_node
    ~landlab.grid.hex.HexModelGrid.closed_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.core_nodes
    ~landlab.grid.hex.HexModelGrid.downwind_links_at_node
    ~landlab.grid.hex.HexModelGrid.fixed_gradient_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.fixed_value_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.link_at_node_is_downwind
    ~landlab.grid.hex.HexModelGrid.link_at_node_is_upwind
    ~landlab.grid.hex.HexModelGrid.link_dirs_at_node
    ~landlab.grid.hex.HexModelGrid.links_at_node
    ~landlab.grid.hex.HexModelGrid.neighbors_at_node
    ~landlab.grid.hex.HexModelGrid.node_at_cell
    ~landlab.grid.hex.HexModelGrid.node_at_core_cell
    ~landlab.grid.hex.HexModelGrid.node_at_link_head
    ~landlab.grid.hex.HexModelGrid.node_at_link_tail
    ~landlab.grid.hex.HexModelGrid.node_axis_coordinates
    ~landlab.grid.hex.HexModelGrid.node_is_boundary
    ~landlab.grid.hex.HexModelGrid.node_x
    ~landlab.grid.hex.HexModelGrid.node_y
    ~landlab.grid.hex.HexModelGrid.nodes
    ~landlab.grid.hex.HexModelGrid.nodes_at_patch
    ~landlab.grid.hex.HexModelGrid.number_of_core_nodes
    ~landlab.grid.hex.HexModelGrid.number_of_links_at_node
    ~landlab.grid.hex.HexModelGrid.number_of_node_columns
    ~landlab.grid.hex.HexModelGrid.number_of_node_rows
    ~landlab.grid.hex.HexModelGrid.number_of_nodes
    ~landlab.grid.hex.HexModelGrid.open_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.patches_at_node
    ~landlab.grid.hex.HexModelGrid.set_nodata_nodes_to_closed
    ~landlab.grid.hex.HexModelGrid.set_nodata_nodes_to_fixed_gradient
    ~landlab.grid.hex.HexModelGrid.status_at_node
    ~landlab.grid.hex.HexModelGrid.unit_vector_sum_xcomponent_at_node
    ~landlab.grid.hex.HexModelGrid.unit_vector_sum_ycomponent_at_node
    ~landlab.grid.hex.HexModelGrid.upwind_links_at_node
    ~landlab.grid.hex.HexModelGrid.x_of_node
    ~landlab.grid.hex.HexModelGrid.y_of_node

Information about links
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.active_link_dirs_at_node
    ~landlab.grid.hex.HexModelGrid.active_links
    ~landlab.grid.hex.HexModelGrid.angle_of_link
    ~landlab.grid.hex.HexModelGrid.downwind_links_at_node
    ~landlab.grid.hex.HexModelGrid.face_at_link
    ~landlab.grid.hex.HexModelGrid.fixed_links
    ~landlab.grid.hex.HexModelGrid.length_of_link
    ~landlab.grid.hex.HexModelGrid.link_at_face
    ~landlab.grid.hex.HexModelGrid.link_at_node_is_downwind
    ~landlab.grid.hex.HexModelGrid.link_at_node_is_upwind
    ~landlab.grid.hex.HexModelGrid.link_dirs_at_node
    ~landlab.grid.hex.HexModelGrid.links_at_node
    ~landlab.grid.hex.HexModelGrid.node_at_link_head
    ~landlab.grid.hex.HexModelGrid.node_at_link_tail
    ~landlab.grid.hex.HexModelGrid.number_of_active_links
    ~landlab.grid.hex.HexModelGrid.number_of_fixed_links
    ~landlab.grid.hex.HexModelGrid.number_of_links
    ~landlab.grid.hex.HexModelGrid.number_of_links_at_node
    ~landlab.grid.hex.HexModelGrid.resolve_values_on_active_links
    ~landlab.grid.hex.HexModelGrid.resolve_values_on_links
    ~landlab.grid.hex.HexModelGrid.status_at_link
    ~landlab.grid.hex.HexModelGrid.unit_vector_xcomponent_at_link
    ~landlab.grid.hex.HexModelGrid.unit_vector_ycomponent_at_link
    ~landlab.grid.hex.HexModelGrid.upwind_links_at_node

Information about cells
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.area_of_cell
    ~landlab.grid.hex.HexModelGrid.cell_area_at_node
    ~landlab.grid.hex.HexModelGrid.cell_at_node
    ~landlab.grid.hex.HexModelGrid.core_cells
    ~landlab.grid.hex.HexModelGrid.faces_at_cell
    ~landlab.grid.hex.HexModelGrid.node_at_cell
    ~landlab.grid.hex.HexModelGrid.node_at_core_cell
    ~landlab.grid.hex.HexModelGrid.number_of_cells
    ~landlab.grid.hex.HexModelGrid.number_of_core_cells
    ~landlab.grid.hex.HexModelGrid.number_of_faces_at_cell

Information about faces
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.active_faces
    ~landlab.grid.hex.HexModelGrid.face_at_link
    ~landlab.grid.hex.HexModelGrid.faces_at_cell
    ~landlab.grid.hex.HexModelGrid.link_at_face
    ~landlab.grid.hex.HexModelGrid.number_of_active_faces
    ~landlab.grid.hex.HexModelGrid.number_of_faces
    ~landlab.grid.hex.HexModelGrid.number_of_faces_at_cell
    ~landlab.grid.hex.HexModelGrid.width_of_face

Information about patches
+++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.nodes_at_patch
    ~landlab.grid.hex.HexModelGrid.number_of_patches
    ~landlab.grid.hex.HexModelGrid.patches_at_node

Data Fields in ModelGrid
------------------------
:class:`~.ModelGrid` inherits from the :class:`~.ModelDataFields` class. This
provides `~.ModelGrid`, and its subclasses, with the ability to, optionally,
store data values that are associated with the different types grid elements
(nodes, cells, etc.). In particular, as part of ``ModelGrid.__init__()``,
data field *groups* are added to the `ModelGrid` that provide containers to
put data fields into. There is one group for each of the eight grid elements
(node, cell, link, face, core_node, core_cell, active_link, and active_face).

To access these groups, use the same methods as accessing groups with
`~.ModelDataFields`. ``ModelGrid.__init__()`` adds the following attributes to
itself that provide access to the values groups:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.grid.hex.HexModelGrid.at_node
    ~landlab.grid.hex.HexModelGrid.at_cell
    ~landlab.grid.hex.HexModelGrid.at_link
    ~landlab.grid.hex.HexModelGrid.at_face

Each of these attributes returns a ``dict``-like object whose keys are value
names as strings and values are numpy arrays that gives quantities at
grid elements.


Create Field Arrays
+++++++++++++++++++
:class:`~.ModelGrid` inherits several useful methods for creating new data
fields and adding new data fields to a ModelGrid instance. Methods to add or
create a new data array follow the ``numpy`` syntax for creating arrays. The
folowing methods create and, optionally, initialize new arrays. These arrays
are of the correct size but a new field will not be added to the field:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.field.grouped.ModelDataFields.empty
    ~landlab.field.grouped.ModelDataFields.ones
    ~landlab.field.grouped.ModelDataFields.zeros

Add Fields to a ModelGrid
+++++++++++++++++++++++++
Unlike with the equivalent numpy functions, these do not take a size argument
as the size of the returned arrays is determined from the size of the
ModelGrid. However, the keyword arguments are the same as those of the numpy
equivalents.

The following methods will create a new array and add a reference to that
array to the ModelGrid:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.grid.hex.HexModelGrid.add_empty
    ~landlab.grid.hex.HexModelGrid.add_field
    ~landlab.grid.hex.HexModelGrid.add_ones
    ~landlab.grid.hex.HexModelGrid.add_zeros
    ~landlab.grid.hex.HexModelGrid.delete_field
    ~landlab.grid.hex.HexModelGrid.set_units

These methods operate in the same way as the previous set except that, in
addition to creating a new array, the newly-created array is added to the
ModelGrid. The calling signature is the same but with the addition of an
argument that gives the name of the new field as a string. The additional
method, :meth:`~.ModelDataFields.add_field`, adds a previously allocation
array to the ModelGrid. If the array is of the incorrect size it will raise
``ValueError``.

Query Fields
++++++++++++
Use the following methods/attributes get information about the stored data
fields:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.field.grouped.ModelDataFields.size
    ~landlab.field.grouped.ModelDataFields.keys
    ~landlab.field.grouped.ModelDataFields.has_group
    ~landlab.field.grouped.ModelDataFields.has_field
    ~landlab.grid.hex.HexModelGrid.field_units
    ~landlab.grid.hex.HexModelGrid.field_values
    ~landlab.field.grouped.ModelDataFields.groups

    i.e., call, e.g. mg.has_field('node', 'my_field_name')

    # START HERE check that all functions listed below are included above,
    # ignore ones that start with underscores(_)

Gradients, fluxes, and divergences on the grid
----------------------------------------------

Landlab is designed to easily calculate gradients in quantities across the
grid, and to construct fluxes and flux divergences from them. Because these
calculations tend to be a little more involved than property lookups, the
methods tend to start with `calc_`.

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.calc_diff_at_link
    ~landlab.grid.hex.HexModelGrid.calc_flux_div_at_node
    ~landlab.grid.hex.HexModelGrid.calc_grad_at_link
    ~landlab.grid.hex.HexModelGrid.calc_grad_at_patch
    ~landlab.grid.hex.HexModelGrid.calc_net_flux_at_node
    ~landlab.grid.hex.HexModelGrid.calc_slope_at_node
    ~landlab.grid.hex.HexModelGrid.calc_slope_at_patch
    ~landlab.grid.hex.HexModelGrid.calc_unit_normal_at_patch

Mappers
-------

These methods allow mapping of values defined on one grid element type onto a
second, e.g., mapping upwind node values onto links, or mean link values onto
nodes.

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.map_downwind_node_link_max_to_node
    ~landlab.grid.hex.HexModelGrid.map_downwind_node_link_mean_to_node
    ~landlab.grid.hex.HexModelGrid.map_link_head_node_to_link
    ~landlab.grid.hex.HexModelGrid.map_link_tail_node_to_link
    ~landlab.grid.hex.HexModelGrid.map_link_vector_to_nodes
    ~landlab.grid.hex.HexModelGrid.map_max_of_link_nodes_to_link
    ~landlab.grid.hex.HexModelGrid.map_max_of_node_links_to_node
    ~landlab.grid.hex.HexModelGrid.map_mean_of_link_nodes_to_link
    ~landlab.grid.hex.HexModelGrid.map_min_of_link_nodes_to_link
    ~landlab.grid.hex.HexModelGrid.map_min_of_node_links_to_node
    ~landlab.grid.hex.HexModelGrid.map_node_to_cell
    ~landlab.grid.hex.HexModelGrid.map_upwind_node_link_max_to_node
    ~landlab.grid.hex.HexModelGrid.map_upwind_node_link_mean_to_node
    ~landlab.grid.hex.HexModelGrid.map_value_at_downwind_node_link_max_to_node
    ~landlab.grid.hex.HexModelGrid.map_value_at_max_node_to_link
    ~landlab.grid.hex.HexModelGrid.map_value_at_min_node_to_link
    ~landlab.grid.hex.HexModelGrid.map_value_at_upwind_node_link_max_to_node

Boundary condition control
--------------------------

These are the primary properties for getting and setting the grid boundary
conditions. Changes made to :meth:`~.ModelGrid.status_at_node` and
:meth:`~.ModelGrid.status_at_node` will automatically update the conditions
defined at other grid elements automatically.

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.active_faces
    ~landlab.grid.hex.HexModelGrid.active_links
    ~landlab.grid.hex.HexModelGrid.boundary_nodes
    ~landlab.grid.hex.HexModelGrid.closed_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.core_cells
    ~landlab.grid.hex.HexModelGrid.core_nodes
    ~landlab.grid.hex.HexModelGrid.fixed_gradient_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.fixed_links
    ~landlab.grid.hex.HexModelGrid.fixed_value_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.node_at_core_cell
    ~landlab.grid.hex.HexModelGrid.node_is_boundary
    ~landlab.grid.hex.HexModelGrid.number_of_active_faces
    ~landlab.grid.hex.HexModelGrid.number_of_active_links
    ~landlab.grid.hex.HexModelGrid.number_of_core_cells
    ~landlab.grid.hex.HexModelGrid.number_of_core_nodes
    ~landlab.grid.hex.HexModelGrid.number_of_fixed_links
    ~landlab.grid.hex.HexModelGrid.open_boundary_nodes
    ~landlab.grid.hex.HexModelGrid.set_nodata_nodes_to_closed
    ~landlab.grid.hex.HexModelGrid.set_nodata_nodes_to_fixed_gradient
    ~landlab.grid.hex.HexModelGrid.status_at_link
    ~landlab.grid.hex.HexModelGrid.status_at_node

Identifying node subsets
------------------------

These methods are useful in identifying subsets of nodes, e.g., closest node
to a point; nodes at edges.

(None are available for this grid type)

Surface analysis
----------------

These methods permit the kinds of surface analysis that you might expect to
find in GIS software.

.. autosummary::
    :toctree: generated/

    ~landlab.grid.hex.HexModelGrid.calc_aspect_at_node
    ~landlab.grid.hex.HexModelGrid.calc_hillshade_at_node
    ~landlab.grid.hex.HexModelGrid.calc_slope_at_node

Notes
-----
It is important that when creating a new grid class that inherits from
``ModelGrid``, to call ``ModelGrid.__init__()`` in the new grid's
``__init__()``. For example, the new class's __init__ should contain the
following code,

.. code-block:: python

    class NewGrid(ModelGrid):
        def __init__(self, *args, **kwds):
            ModelGrid.__init__(self, **kwds)
            # Code that initializes the NewGrid

Without this, the new grid class will not have the ``at_*`` attributes.
"""

import numpy
import six

from landlab.grid.voronoi import VoronoiDelaunayGrid


class HexModelGrid(VoronoiDelaunayGrid):
    """A grid of hexagonal cells.

    This inherited class implements a regular 2D grid with hexagonal cells and
    triangular patches. It is a special type of VoronoiDelaunay grid in which
    the initial set of points is arranged in a triangular/hexagonal lattice.

    Parameters
    ----------
    base_num_rows : int
        Number of rows of nodes in the left column.
    base_num_cols : int
        Number of nodes on the first row.
    dx : float, optional
        Node spacing.
    orientation : string, optional
        One of the 3 cardinal directions in the grid, either 'horizontal'
        (default) or 'vertical'

    Returns
    -------
    HexModelGrid
        A newly-created grid.

    Examples
    --------
    Create a hex grid with 2 rows of nodes. The first and third rows will
    have 2 nodes, and the second nodes.

    >>> from landlab import HexModelGrid
    >>> hmg = HexModelGrid(3, 2, 1.0)
    >>> hmg.number_of_nodes
    7
    """

    def __init__(self, base_num_rows=0, base_num_cols=0, dx=1.0,
                 orientation='horizontal', shape='hex', reorient_links=True,
                 **kwds):
        """Create a grid of hexagonal cells.

        Create a regular 2D grid with hexagonal cells and triangular patches.
        It is a special type of VoronoiDelaunay grid in which the initial set
        of points is arranged in a triangular/hexagonal lattice.

        Parameters
        ----------
        base_num_rows : int
            Number of rows of nodes in the left column.
        base_num_cols : int
            Number of nodes on the first row.
        dx : float, optional
            Node spacing.
        orientation : string, optional
            One of the 3 cardinal directions in the grid, either 'horizontal'
            (default) or 'vertical'

        Returns
        -------
        HexModelGrid
            A newly-created grid.

        Examples
        --------
        Create a hex grid with 2 rows of nodes. The first and third rows will
        have 2 nodes, and the second nodes.

        >>> from landlab import HexModelGrid
        >>> hmg = HexModelGrid(3, 2, 1.0)
        >>> hmg.number_of_nodes
        7
        """
        # Set number of nodes, and initialize if caller has given dimensions
        if base_num_rows * base_num_cols > 0:
            self._initialize(base_num_rows, base_num_cols, dx, orientation,
                             shape, reorient_links)
        super(HexModelGrid, self).__init__(**kwds)

    @classmethod
    def from_dict(cls, params):
        shape = params['shape']
        spacing = params.get('spacing', 1.)

        return cls(shape[0], shape[1], spacing)

    def _initialize(self, base_num_rows, base_num_cols, dx, orientation,
                    shape, reorient_links=True):
        r"""Set up a hexagonal grid.

        Sets up a hexagonal grid with cell spacing dx and
        (by default) regular boundaries (that is, all perimeter cells are
        boundaries and all interior cells are active).

        Parameters
        ----------
        base_num_rows : int
            Number of rows along left side of grid
        base_num_cols : int
            Number of columns along bottom side of grid
        dx : float
            Distance between nodes
        orientation : string
            Either 'horizontal' (default in __init__) or 'vertical'
        shape : string
            Either 'hex' (default in __init__) or 'rect'
        reorient_links : bool
            Whether or not to re-orient all links to point between -45 deg
            and +135 deg clockwise from "north" (i.e., along y axis)

        Returns
        -------
        (none)

        Creates/modifies
        ----------------
        Creates and initializes and self._dx

        Notes
        -----
        To be consistent with unstructured grids, the hex grid is
        managed not as a 2D array but rather as a set of arrays that
        describe connectivity information between nodes, links, cells, faces,
        patches, corners, and junctions.

        'Horizontal' orientation means that one of the 3 axes of the grid is
        horizontal, whereas the other two are at 30 degree angles to the
        horizontal, like:

            \ /
           -----
            / \

        'Vertical' means that one axis is vertical, with the other
        two at 30 degree angles to the vertical, more like:

           \   |   /
             \ | /
             / | \
           /   |   \

        (of course, these keyboard characters don't represent the angles quite
        right)

        Numbers of rows and columns: a hex grid with a rectangular shape will
        have a fixed number of rows and columns, and so for rectangular shaped
        grids we record this information in self._nrows and self._ncols. With
        a hex-shaped grid, either the number of columns (if 'horizontal') or
        the number of rows (if 'vertical') will vary across the grid.
        Therefore, for hex-shaped grids we record only self._nrows for
        'horizontal' grids, and only self._ncols for 'vertical' grids.
        """
        if self._DEBUG_TRACK_METHODS:
            six.print_('HexModelGrid._initialize(' + str(base_num_rows) +
                       ', ' + str(base_num_cols) + ', ' + str(dx) + ')')

        # Make sure the parameter *orientation* is correct
        assert (orientation[0].lower() == 'h' or
                orientation[0].lower() == 'v'), \
            'orientation must be either "horizontal" (default) or "vertical"'

        # Make sure the parameter *shape* is correct
        assert (shape[0].lower() == 'h' or shape[0].lower() == 'r'), \
            'shape must be either "hex" (default) or "rect"'

        # Create a set of hexagonally arranged points. These will be our nodes.
        if orientation == 'horizontal' and shape == 'hex':
            pts = HexModelGrid._hex_points_with_horizontal_hex(
                base_num_rows, base_num_cols, dx)
            self.orientation = 'horizontal'
            self._nrows = base_num_rows
        elif orientation == 'horizontal' and shape == 'rect':
            pts = HexModelGrid._hex_points_with_horizontal_rect(
                base_num_rows, base_num_cols, dx)
            self.orientation = 'horizontal'
            self._nrows = base_num_rows
            self._ncols = base_num_cols
        elif orientation == 'vertical' and shape == 'hex':
            pts = HexModelGrid._hex_points_with_vertical_hex(
                base_num_rows, base_num_cols, dx)
            self.orientation = 'vertical'
            self._ncols = base_num_cols
        else:
            pts = HexModelGrid._hex_points_with_vertical_rect(
                base_num_rows, base_num_cols, dx)
            self.orientation = 'vertical'
            self._nrows = base_num_rows
            self._ncols = base_num_cols

        # Call the VoronoiDelaunayGrid constructor to triangulate/Voronoi
        # the nodes into a grid.
        super(HexModelGrid, self)._initialize(
            pts[:, 0], pts[:, 1], reorient_links)

        # Remember grid spacing
        self._dx = dx

    def _create_cell_areas_array(self):
        r"""Create an array of surface areas of hexagonal cells.

        Creates and returns an array containing the surface areas of the
        hexagonal (Voronoi) cells.

        These cells are perfect hexagons in which the apothem is dx/2. The
        formula for area is:

        .. math::
            A = 3 dx^2 / 2 \sqrt{3} \approx 0.866 dx^2
        """
        self._area_of_cell = (0.8660254 * self._dx ** 2 +
                              numpy.zeros(self.number_of_cells))
        return self._area_of_cell

    @staticmethod
    def _hex_points_with_horizontal_hex(num_rows, base_num_cols, dxh):
        """Create a set of points on a staggered grid.

        Creates and returns a set of (x,y) points in a staggered grid in which
        the points represent the centers of regular hexagonal cells, and the
        points could be connected to form equilateral triangles. The overall
        shape of the lattice is hexagonal, and one of the 3 axes is horizontal.

        Parameters
        ----------
        num_rows : int
            Number of rows in lattice
        base_num_cols : int
            Number of columns in the bottom and top rows (middle rows have
            more)
        dxh : float
            Horizontal and diagonal spacing between points

        Returns
        -------
        poinst : ndarray
            A 2D numpy array containing point (x,y) coordinates, and total
            number of points.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> points = HexModelGrid._hex_points_with_horizontal_hex(3, 2, 1.0)
        >>> len(points)
        7
        >>> points[1, :]
        array([ 1.,  0.])
        >>> points[:3, 0]
        array([ 0. ,  1. , -0.5])
        """
        dxv = dxh * numpy.sqrt(3.) / 2.
        half_dxh = dxh / 2.

        if numpy.mod(num_rows, 2) == 0:  # even number of rows
            npts = num_rows * base_num_cols + (num_rows * num_rows) // 4
        else:  # odd number of rows
            npts = num_rows * base_num_cols + \
                ((num_rows - 1) // 2) * ((num_rows - 1) // 2)
        pts = numpy.zeros((npts, 2))
        middle_row = num_rows // 2
        extra_cols = 0
        xshift = 0.
        i = 0
        for r in range(num_rows):
            for c in range(base_num_cols + extra_cols):
                pts[i, 0] = c * dxh + xshift
                pts[i, 1] = r * dxv
                i += 1
            if r < middle_row:
                extra_cols += 1
            else:
                extra_cols -= 1
            xshift = - half_dxh * extra_cols

        return pts

    @staticmethod
    def _hex_points_with_horizontal_rect(num_rows, num_cols, dxh):
        """Create a set of points in a taggered grid.
        Creates and returns a set of (x,y) points in a staggered grid in which
        the points represent the centers of regular hexagonal cells, and the
        points could be connected to form equilateral triangles. The overall
        shape of the lattice is rectangular, and one of the 3 axes is
        horizontal.

        Parameters
        ----------
        num_rows : int
            Number of rows in lattice
        num_cols : int
            Number of columns in lattice
        dxh : float
            Horizontal and diagonal spacing between points

        Returns
        -------
        points : ndarray of shape `(n_points, 2)`
            A 2D numpy array containing point (x, y) coordinates, and total
            number of points.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> points = HexModelGrid._hex_points_with_horizontal_rect(3, 3, 1.0)
        >>> len(points)
        9
        >>> points[1, :]
        array([ 1.,  0.])
        >>> points[:3, 0]
        array([ 0.,  1.,  2.])
        """
        dxv = dxh * numpy.sqrt(3.) / 2.
        half_dxh = dxh / 2.

        npts = num_rows * num_cols
        pts = numpy.zeros((npts, 2))
        xshift = 0.
        i = 0
        for r in range(num_rows):
            for c in range(num_cols):
                xshift = half_dxh * (r % 2)
                pts[i, 0] = c * dxh + xshift
                pts[i, 1] = r * dxv
                i += 1

        return pts

    @staticmethod
    def _hex_points_with_vertical_hex(base_num_rows, num_cols, dxv):
        """
        Creates and returns a set of (x,y) points in a staggered grid in which
        the points represent the centers of regular hexagonal cells, and the
        points could be connected to form equilateral triangles. The overall
        shape of the lattice is hexagonal.

        Parameters
        ----------
        base_num_rows : int
            Number of columns in the left and right columns (middle columns
            have more)
        num_cols : int
            Number of columns in lattice
        dxv : float
            Vertical and diagonal spacing between points

        Returns
        -------
        points : ndarray of shape `(n_points, 2)`
            2D numpy array containing point (x,y) coordinates, and total
            number of points.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> points = HexModelGrid._hex_points_with_vertical_hex(2, 3, 1.0)
        >>> len(points)
        7
        >>> points[1, :]
        array([ 0.,  1.])
        >>> points[:3, 1]
        array([ 0. ,  1. , -0.5])
        """
        dxh = dxv * numpy.sqrt(3.) / 2.
        half_dxv = dxv / 2.

        if numpy.mod(num_cols, 2) == 0:  # even number of columns
            npts = base_num_rows * num_cols + (num_cols * num_cols) // 4
        else:  # odd number of columns
            npts = base_num_rows * num_cols + \
                ((num_cols - 1) // 2) * ((num_cols - 1) // 2)
        pts = numpy.zeros((npts, 2))
        middle_col = num_cols // 2
        extra_rows = 0
        yshift = 0.
        i = 0
        for c in range(num_cols):
            for r in range(base_num_rows + extra_rows):
                pts[i, 1] = r * dxv + yshift
                pts[i, 0] = c * dxh
                i += 1
            if c < middle_col:
                extra_rows += 1
            else:
                extra_rows -= 1
            yshift = - half_dxv * extra_rows

        return pts

    @staticmethod
    def _hex_points_with_vertical_rect(num_rows, num_cols, dxv):
        """
        Creates and returns a set of (x,y) points in a staggered grid in which
        the points represent the centers of regular hexagonal cells, and the
        points could be connected to form equilateral triangles. The overall
        shape of the lattice is rectangular.

        Parameters
        ----------
        num_rows : int
            Number of columns in lattice
        num_cols : int
            Number of columns in lattice
        dxv : float
            Vertical and diagonal spacing between points

        Returns
        -------
        points : ndarray of shape `(n_points, 2)`
            2D numpy array containing point (x,y) coordinates, and total
            number of points.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> points = HexModelGrid._hex_points_with_vertical_rect(3, 3, 1.0)
        >>> len(points)
        9
        >>> points[1, :]
        array([ 0.,  1.])
        >>> points[:3, 1]
        array([ 0.,  1.,  2.])
        """
        dxh = dxv * numpy.sqrt(3.) / 2.
        half_dxv = dxv / 2.

        npts = num_rows * num_cols
        pts = numpy.zeros((npts, 2))
        yshift = 0.
        i = 0
        for c in range(num_cols):
            for r in range(num_rows):
                yshift = half_dxv * (c % 2)
                pts[i, 1] = r * dxv + yshift
                pts[i, 0] = c * dxh
                i += 1

        return pts

    @property
    def number_of_node_columns(self):
        """Number of node columns hex grid.

        Number of node columns in a rectangular-shaped and/or
        vertically oriented hex grid.

        Returns the number of columns, including boundaries.

        Notes
        -----
        Will generate an error if called with a hex-shaped, horizontally
        aligned grid.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> grid = HexModelGrid(5, 5, shape='rect')
        >>> grid.number_of_node_columns
        5
        """
        return self._ncols

    @property
    def number_of_node_rows(self):
        """Number of node rows in a rectangular-shaped and/or
        horizontally oriented hex grid.

        Returns the number of rows, including boundaries.

        Notes
        -----
        Will generate an error if called with a hex-shaped, vertically
        aligned grid.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> grid = HexModelGrid(5, 5, shape='rect')
        >>> grid.number_of_node_rows
        5
        """
        return self._nrows

    def _configure_hexplot(self, data, data_label=None, color_map=None):
        """
        Sets up necessary information for making plots of the hexagonal grid
        colored by a given data element.

        Parameters
        ----------
        data : str OR node array (1d numpy array with number_of_nodes entries)
            Data field to be colored
        data_label : str, optional
            Label for colorbar
        color_map : matplotlib colormap object, None
            Color map to apply (defaults to "jet")

        Returns
        -------
        (none)

        Notes
        -----
        Creates and stores a PatchCollection representing the hexagons. Also
        stores a handle to the current plotting axis. Both of these are then
        used by hexplot().
        """
        from numpy import array, sqrt, zeros
        import matplotlib
        from matplotlib.patches import Polygon
        from matplotlib.collections import PatchCollection

        # color
        if color_map is None:
            color_map = matplotlib.cm.jet

        # geometry
        apothem = self._dx / 2.0
        # distance from node to each hexagon cell vertex
        radius = 2.0 * apothem / sqrt(3.0)

        # offsets from node x,y position
        offsets = zeros((6, 2))
        poly_verts = zeros((6, 2))

        # Figure out whether the orientation is horizontal or vertical
        if self.orientation == 'horizontal':   # horizontal
            offsets[:, 0] = array(
                [0., apothem, apothem, 0., -apothem, -apothem])
            offsets[:, 1] = array(
                [radius, radius / 2.0, -radius / 2.0, -radius, -radius / 2.0,
                 radius / 2.0])
        else:   # vertical
            offsets[:, 0] = array(
                [radius / 2.0, radius, radius / 2.0, -radius / 2.0, -radius,
                 -radius / 2.0])
            offsets[:, 1] = array(
                [apothem, 0., -apothem, -apothem, 0., apothem])

        patches = []
        for i in range(self.number_of_nodes):
            poly_verts[:, 0] = self.node_x[i] + offsets[:, 0]
            poly_verts[:, 1] = self.node_y[i] + offsets[:, 1]
            p = Polygon(poly_verts, True)
            patches.append(p)

        self._hexplot_pc = PatchCollection(
            patches, cmap=color_map, edgecolor='none', linewidth=0.0)

        self._hexplot_configured = True

    def hexplot(self, data, data_label=None, color_map=None):
        """Create a plot of the grid elements.

        Creates a plot of the grid and one node-data field, showing hexagonal
        cells colored by values in the field.

        Parameters
        ----------
        data : str or node array (1d numpy array with number_of_nodes entries)
            Data field to be colored.
        data_label : str, optional
            Label for colorbar.
        color_map : matplotlib colormap object, None
            Color map to apply (defaults to "jet")

        See also
        --------
        plot.imshow_grid
            Another Landlab function capable of producing hexplots, with a
            fuller-featured set of options.
        """
        from numpy import array, amin, amax
        import matplotlib.pyplot as plt
        import copy

        try:
            self._hexplot_configured is True
        except:
            self._configure_hexplot(data, data_label, color_map)

        # Handle *data*: if it's a numpy array, then we consider it the
        # data to be plotted. If it's a string, we consider it the name of the
        # node-field to plot, and we fetch it.
        if type(data) is str:
            data_label = data
            data = self.at_node[data]

        ax = plt.gca()
        self._hexplot_pc.set_array(array(data))
        copy_of_pc = copy.copy(self._hexplot_pc)
        ax.add_collection(copy_of_pc)
        plt.xlim([amin(self.node_x) - self._dx, amax(self.node_x) + self._dx])
        plt.ylim([amin(self.node_y) - self._dx, amax(self.node_y) + self._dx])

        return ax


def from_dict(param_dict):
    """
    Create a HexModelGrid from the dictionary-like object, *param_dict*.
    Required keys of the dictionary are NUM_ROWS, NUM_COLS. Raises a KeyError
    if either of these are missing.  If GRID_SPACING is given, use it as the
    HexModelGrid *dx* parameter, otherwise default to unit spacing.
    """
    # Read and create a basic HexModelGrid
    try:
        n_rows = int(param_dict['NUM_ROWS'])
        n_cols = int(param_dict['NUM_COLS'])
        dx = float(param_dict.get('GRID_SPACING', 1.))
    except KeyError as e:
        raise
    except ValueError as e:
        raise
    else:
        hg = HexModelGrid(n_rows, n_cols, dx)

    return hg



#! /usr/bin/env python
"""Store collections of data fields."""

from .scalar_data_fields import ScalarDataFields


class Error(Exception):

    """Base class for errors in this module."""

    pass


class GroupError(Error, KeyError):

    """Raise this error for a missing group name."""

    def __init__(self, group):
        self._group = group

    def __str__(self):
        return self._group


class GroupSizeError(Error, KeyError):

    """Raise this error if a group has changed sizes."""

    def __init__(self, group, old_size, new_size):
        self._group = group
        self._old_size = old_size
        self._new_size = new_size

    def __str__(self):
        return ("number of {group} elements has changed. "
                "(was = {was}, now={now})".format(group=self._group,
                                                  was=self._old_size,
                                                  now=self._new_size))


class ModelDataFields(object):

    """Collection of grouped data-fields.

    The ModelDataFields class holds a set of ScalarDataFields that are
    separated into *groups*. A typical use for this class would be to define
    the groups as being locations on a grid where the values are defined.
    For instance, the groups could be *node*, *cell*, *link*, and *face*.

    Most of the method functions for ModelDataFields are the same as those for
    the ScalarDataFields class but with the first argument being a string that
    defines the group name.

    Attributes
    ----------
    groups

    See Also
    --------
    landlab.field.ScalarDataFields : Data fields within a *group* are
        stored as :class:`landlab.field.ScalarDataFields`.
    landlab.grid.ModelGrid : Inherits from ModelDataFields.

    Examples
    --------

    Create two groups of data fields defined at *node* and *cell*. Each set can
    have a differenct number of values.

    >>> from landlab.field import ModelDataFields
    >>> fields = ModelDataFields()
    >>> fields.new_field_location('node', 12)
    >>> fields.new_field_location('cell', 2)

    Create some new value arrays for each of the data fields.

    >>> fields.ones('node')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])
    >>> fields.zeros('cell')
    array([ 0.,  0.])

    Create new value arrays and add them to the data fields. Because the data
    fields are in different groups (node and cell), they can have the same
    name.

    >>> fields.add_ones('node', 'topographic__elevation')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])
    >>> fields.at_node['topographic__elevation']
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])

    >>> fields.add_ones('cell', 'topographic__elevation')
    array([ 1.,  1.])
    >>> fields.at_cell['topographic__elevation']
    array([ 1.,  1.])

    Each group acts as a `dict` so, for instance, to get the variables names
    in a group use the `keys` method,

    >>> list(fields.at_cell.keys())
    ['topographic__elevation']
    """

    def __init__(self, **kwds):
        self._groups = dict()
        super(ModelDataFields, self).__init__(**kwds)

    @property
    def groups(self):
        """List of group names.

        Returns
        -------
        set
            Set of quantity names.
        """
        return set(self._groups.keys())

    def has_group(self, group):
        """Check if a group exists.

        Parameters
        ----------
        group: str
            Name of the group.

        Returns
        -------
        boolean
            True if the field contains *group*, otherwise False.

        Examples
        --------
        Check if the field has the groups named *node* or *cell*.

        >>> from landlab.field import ModelDataFields
        >>> fields = ModelDataFields()
        >>> fields.new_field_location('node', 12)
        >>> fields.has_group('node')
        True
        >>> fields.has_group('cell')
        False
        """
        return group in self._groups

    def has_field(self, group, field):
        """Check if a field is in a group.

        Parameters
        ----------
        group: str
            Name of the group.
        field: str
            Name of the field.

        Returns
        -------
        boolean
            ``True`` if the group contains the field, otherwise ``False``.

        Examples
        --------
        Check if the field named ``topographic__elevation`` is contained
        in a group.

        >>> from landlab.field import ModelDataFields
        >>> fields = ModelDataFields()
        >>> fields.new_field_location('node', 12)
        >>> _ = fields.add_ones('node', 'topographic__elevation')
        >>> fields.has_field('node', 'topographic__elevation')
        True
        >>> fields.has_field('cell', 'topographic__elevation')
        False
        """
        try:
            return field in self[group]
        except KeyError:
            return False

    def keys(self, group):
        """List of field names in a group.

        Returns a list of the field names as a list of strings.

        Parameters
        ----------
        group : str
            Group name.

        Returns
        -------
        list
            List of field names.

        Examples
        --------
        >>> from landlab.field import ModelDataFields
        >>> fields = ModelDataFields()
        >>> fields.new_field_location('node', 4)
        >>> list(fields.keys('node'))
        []
        >>> _ = fields.add_empty('node', 'topographic__elevation')
        >>> list(fields.keys('node'))
        ['topographic__elevation']
        """
        return self[group].keys()

    def size(self, group):
        """Size of the arrays stored in a group.

        Parameters
        ----------
        group : str
            Group name.

        Returns
        -------
        int
            Array size.

        Examples
        --------
        >>> from landlab.field import ModelDataFields
        >>> fields = ModelDataFields()
        >>> fields.new_field_location('node', 4)
        >>> fields.size('node')
        4
        """
        return self[group].size

    def new_field_location(self, group, size=None):
        """Add a new quantity to a field.

        Create an empty group into which new fields can be added. The new group
        is created but no memory allocated yet. The dictionary of the new group
        can be through a new *at_* attribute of the class instance.

        Parameters
        ----------
        group: str
            Name of the new group to add to the field.
        size: int, optional
            Number of elements in the new quantity. If not provided, the
            size is set to be the size of the first field added to the goup.

        Raises
        ------
        ValueError
            If the field already contains the group.

        Examples
        --------
        Create a collection of fields and add two groups, *node* and *cell*,
        to it.

        >>> from landlab.field import ModelDataFields
        >>> fields = ModelDataFields()
        >>> fields.new_field_location('node', 12)
        >>> fields.new_field_location('cell', 2)

        The group names in the collection are retrieved with the *groups*
        attribute as a `set`.

        >>> names = list(fields.groups)
        >>> names.sort()
        >>> names
        ['cell', 'node']

        Access the new (empty) groups with the *at_* attributes.

        >>> fields.at_cell, fields.at_node
        ({}, {})

        >>> fields.new_field_location('core_node')
        >>> fields.at_core_node.size is None
        True
        >>> fields.at_core_node['air__temperature'] = [0, 1]
        >>> fields.at_core_node.size
        2
        """
        if self.has_group(group):
            raise ValueError('ModelDataFields already contains %s' % group)
        else:
            self._groups[group] = ScalarDataFields(size)
            setattr(self, 'at_' + group, self[group])

    def field_values(self, group, field):
        """Get values of a field.

        Given a *group* and a *field*, return a reference to the associated
        data array.

        Parameters
        ----------
        group: str
            Name of the group.
        field: str
            Name of the field withing *group*.

        Returns
        -------
        array
            The values of the field.

        Raises
        ------
        GroupError
            If *group* does not exits
        FieldError
            If *field* does not exits

        Examples
        --------
        Create a group of fields called *node*.

        >>> from landlab.field import ModelDataFields
        >>> fields = ModelDataFields()
        >>> fields.new_field_location('node', 4)

        Add a field, initialized to ones, called *topographic__elevation*
        to the *node* group. The *field_values* method returns a reference
        to the field's data.

        >>> _ = fields.add_ones('node', 'topographic__elevation')
        >>> fields.field_values('node', 'topographic__elevation')
        array([ 1.,  1.,  1.,  1.])

        Raise FieldError if *field* does not exist in *group*.

        >>> fields.field_values('node', 'planet_surface__temperature')
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: planet_surface__temperature

        If *group* does not exists, Raise GroupError.

        >>> fields.field_values('cell', 'topographic__elevation')
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        GroupError: cell
        """
        return self[group][field]

    def field_units(self, group, field):
        """Get units for a field.

        Returns the unit string associated with the data array in *group* and
        *field*.

        Parameters
        ----------
        group: str
            Name of the group.
        field: str
            Name of the field withing *group*.

        Returns
        -------
        str
            The units of the field.

        Raises
        ------
        KeyError
            If either *field* or *group* does not exist.
        """
        return self[group].units[field]

    def empty(self, group, **kwds):
        """Uninitialized array whose size is that of the field.

        Return a new array of the data field size, without initializing
        entries. Keyword arguments are the same as that for the equivalent
        numpy function.

        Parameters
        ----------
        group : str
            Name of the group.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.ModelDataFields.ones : Equivalent method that
            initializes the data to 1.
        landlab.field.ModelDataFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import ModelDataFields
        >>> field = ModelDataFields()
        >>> field.new_field_location('node', 4)
        >>> field.empty('node') # doctest: +SKIP
        array([  2.31584178e+077,  -2.68156175e+154,   9.88131292e-324,
        ... 2.78134232e-309]) # Uninitialized memory

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys('node'))
        []
        """
        return self[group].empty(**kwds)

    def ones(self, group, **kwds):
        """Array, initialized to 1, whose size is that of the field.

        Return a new array of the data field size, filled with ones. Keyword
        arguments are the same as that for the equivalent numpy function.

        Parameters
        ----------
        group : str
            Name of the group.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        landlab.field.ModelDataFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ModelDataFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import ModelDataFields
        >>> field = ModelDataFields()
        >>> field.new_field_location('node', 4)
        >>> field.ones('node')
        array([ 1.,  1.,  1.,  1.])
        >>> field.ones('node', dtype=int)
        array([1, 1, 1, 1])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys('node'))
        []
        """
        return self[group].ones(**kwds)

    def zeros(self, group, **kwds):
        """Array, initialized to 0, whose size is that of the field.

        Parameters
        ----------
        group : str
            Name of the group.

        Return a new array of the data field size, filled with zeros. Keyword
        arguments are the same as that for the equivalent numpy function.

        See Also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.ModelDataFields.empty : Equivalent method that does not
            initialize the new array.
        landlab.field.ModelDataFields.ones : Equivalent
            method that initializes the data to 1.

        Examples
        --------
        >>> from landlab.field import ModelDataFields
        >>> field = ModelDataFields()
        >>> field.new_field_location('node', 4)
        >>> field.zeros('node')
        array([ 0.,  0.,  0.,  0.])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys('node'))
        []
        """
        return self[group].zeros(**kwds)

    def add_empty(self, *args, **kwds):
        """
        Create and add an uninitialized array of values to the field.

        Create a new array of the data field size, without initializing
        entries, and add it to the field as *name*. The *units* keyword gives
        the units of the new fields as a string. Remaining keyword arguments
        are the same as that for the equivalent numpy function.

        Construction::

            add_empty(group, name, units='-', noclobber=True)

        Parameters
        ----------
        group : str
            Name of the group.
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.ModelDataFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ModelDataFields.zeros : Equivalent method that
            initializes the data to 0.
        """
        if len(args) == 2:
            group, name = args
        elif len(args) == 1:
            group, name = kwds.pop('at'), args[0]
        else:
            raise ValueError('number of arguments must be 1 or 2')

        numpy_kwds = kwds.copy()
        numpy_kwds.pop('units', 0.)
        numpy_kwds.pop('noclobber', 0.)
        return self.add_field(group, name, self.empty(group, **numpy_kwds),
                              **kwds)

    def add_ones(self, *args, **kwds):
        """
        Create and add an array of values, initialized to 1, to the field.

        Create a new array of the data field size, filled with ones, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.

        Construction::

            add_ones(group, name, units='-', noclobber=True)

        Parameters
        ----------
        group : str
            Name of the group.
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        andlab.field.ModelDataFields.add_empty : Equivalent method that
            does not initialize the new array.
        andlab.field.ModelDataFields.add_zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        Add a new, named field to a collection of fields.

        >>> from landlab.field import ModelDataFields
        >>> field = ModelDataFields()
        >>> field.new_field_location('node', 4)
        >>> field.add_ones('node', 'topographic__elevation')
        array([ 1.,  1.,  1.,  1.])
        >>> list(field.keys('node'))
        ['topographic__elevation']
        >>> field['node']['topographic__elevation']
        array([ 1.,  1.,  1.,  1.])
        >>> field.at_node['topographic__elevation']
        array([ 1.,  1.,  1.,  1.])
        """
        if len(args) == 2:
            group, name = args
        elif len(args) == 1:
            group, name = kwds.pop('at'), args[0]
        else:
            raise ValueError('number of arguments must be 1 or 2')

        numpy_kwds = kwds.copy()
        numpy_kwds.pop('units', 0.)
        numpy_kwds.pop('noclobber', 0.)
        return self.add_field(group, name, self.ones(group, **numpy_kwds),
                              **kwds)

    def add_zeros(self, *args, **kwds):
        """
        Create and add an array of values, initialized to 0, to the field.

        Create a new array of the data field size, filled with zeros, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.

        Construction::

            add_zeros(group, name, units='-', noclobber=True)

        Parameters
        ----------
        group : str
            Name of the group.
        name : str
            Name of the new field to add.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.ScalarDataFields.add_empty : Equivalent method that
            does not initialize the new array.
        landlab.field.ScalarDataFields.add_ones : Equivalent method that
            initializes the data to 1.
        """
        if len(args) == 2:
            group, name = args
        elif len(args) == 1:
            group, name = kwds.pop('at'), args[0]
        else:
            raise ValueError('number of arguments must be 1 or 2')

        numpy_kwds = kwds.copy()
        numpy_kwds.pop('units', 0.)
        numpy_kwds.pop('noclobber', 0.)
        return self.add_field(group, name, self.zeros(group, **numpy_kwds),
                              **kwds)

    def add_field(self, *args, **kwds):
        """Add an array of values to the field.

        Add an array of data values to a collection of fields and associate it
        with the key, *name*. Use the *copy* keyword to, optionally, add a
        copy of the provided array.

        Construction::

            add_field(group, name, value_array, units='-', copy=False,
                      noclobber=True)

        Parameters
        ----------
        group : str
            Name of the group.
        name : str
            Name of the new field to add.
        value_array : numpy.array
            Array of values to add to the field.
        units : str, optional
            Optionally specify the units of the field.
        copy : boolean, optional
            If True, add a *copy* of the array to the field. Otherwise save add
            a reference to the array.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.array
            The data array added to the field. Depending on the *copy*
            keyword, this could be a copy of *value_array* or *value_array*
            itself.

        Raises
        ------
        ValueError :
            If *value_array* has a size different from the field.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.field import ModelDataFields
        >>> field = ModelDataFields()
        >>> field.new_field_location('node', 4)
        >>> values = np.ones(4, dtype=int)
        >>> field.add_field('node', 'topographic__elevation', values)
        array([1, 1, 1, 1])

        A new field is added to the collection of fields. The saved value
        array is the same as the one initially created.

        >>> field.at_node['topographic__elevation'] is values
        True

        If you want to save a copy of the array, use the *copy* keyword. In
        addition, adding values to an existing field will remove the reference
        to the previously saved array. The *noclobber* keyword changes this
        behavior to raise an exception in such a case.

        >>> field.add_field('node', 'topographic__elevation', values,
        ...     copy=True, noclobber=False)
        array([1, 1, 1, 1])
        >>> field.at_node['topographic__elevation'] is values
        False
        >>> field.add_field('node', 'topographic__elevation', values,
        ...     noclobber=True) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: topographic__elevation
        """
        if len(args) == 3:
            group, name, value_array = args
        elif len(args) == 2:
            group, name, value_array = kwds.pop('at'), args[0], args[1]
        else:
            raise ValueError('number of arguments must be 2 or 3')

        return self[group].add_field(name, value_array, **kwds)

    def set_units(self, group, name, units):
        """Set the units for a field of values.

        Parameters
        ----------
        group : str
            Name of the group.
        name: str
            Name of the field.
        units: str
            Units for the field.

        Raises
        ------
        KeyError
            If the named field does not exist.
        """
        self[group].set_units(name, units)

    def delete_field(self, group, name):
        """Erases an existing field.

        Parameters
        ----------
        group : str
            Name of the group.
        name: str
            Name of the field.

        Raises
        ------
        KeyError
            If the named field does not exist.
        """
        del self._groups[group].units[name]
        del self._groups[group][name]

    def __getitem__(self, group):
        """Get a group of fields."""
        try:
            return self._groups[group]
        except KeyError:
            raise GroupError(group)



import numpy as np

from numpy.testing import assert_array_equal
from nose.tools import raises, assert_equal

from landlab.grid.structured_quad.nodes import status_with_perimeter_as_boundary
from landlab.grid.structured_quad.links import active_link_ids

from landlab.grid.base import CORE_NODE, FIXED_VALUE_BOUNDARY, CLOSED_BOUNDARY
from landlab.testing.tools import assert_array_is_int


def test_active_links_ids():
    status = np.empty((4, 5), dtype=int)
    status.fill(CLOSED_BOUNDARY)
    status[1, 2] = status[1, 3] = status[2, 2] = status[2, 3] = CORE_NODE

    link_ids = active_link_ids((4, 5), status)
    assert_array_equal(link_ids, [11, 15, 16, 20])
    # assert_array_equal(link_ids, [7, 8, 21, 25])
    assert_array_is_int(link_ids)


def test_active_links_with_edge_boundaries():
    status = status_with_perimeter_as_boundary((3, 4))
    link_ids = active_link_ids((3, 4), status)
    assert_array_equal(link_ids, [8])
    assert_array_is_int(link_ids)


@raises(ValueError)
def test_active_link_ids_with_shape_mismatch():
    active_link_ids((3, 4), np.zeros(3))



import numpy as np
from numpy.testing import assert_array_equal
from nose import with_setup
from nose.tools import (assert_equal, assert_raises)
try:
    from nose.tools import assert_is
except ImportError:
    from landlab.testing.tools import assert_is

from landlab.grid import raster_funcs as rfuncs
from landlab import RasterModelGrid


def test_with_arrays():
    """Test with arrays as arg."""
    rmg = RasterModelGrid((4, 5), spacing=(2., 2.))

    coords = (np.array([1., -1.]), np.array([1., -1.]))
    assert_array_equal(rfuncs.is_coord_on_grid(rmg, coords),
                       np.array([True, False]))


def test_just_inside():
    """Test with points just inside the grid."""
    rmg = RasterModelGrid((4, 5), spacing=(2., 2.))

    assert_equal(rfuncs.is_coord_on_grid(rmg, (0., 4.)), True)
    assert_equal(rfuncs.is_coord_on_grid(rmg, (8. - 1e-12, 4.)), True)
    assert_equal(rfuncs.is_coord_on_grid(rmg, (3., 0.)), True)
    assert_equal(rfuncs.is_coord_on_grid(rmg, (3., 6. - 1e-12)), True)


def test_just_outside():
    """Test with points just outside the grid."""
    rmg = RasterModelGrid((4, 5), spacing=(2., 2.))

    assert_equal(rfuncs.is_coord_on_grid(rmg, (0. - 1e-12, 4.)), False)
    assert_equal(rfuncs.is_coord_on_grid(rmg, (8., 4.)), False)
    assert_equal(rfuncs.is_coord_on_grid(rmg, (3., 0. - 1e-12)), False)
    assert_equal(rfuncs.is_coord_on_grid(rmg, (3., 6.)), False)


def test_just_x():
    """Test check if points are within the x bounds."""
    rmg = RasterModelGrid((4, 5), spacing=(2., 2.))
    assert_equal(rfuncs.is_coord_on_grid(rmg, (4., 1.e6), axes=(1, )), True)
    assert_equal(rfuncs.is_coord_on_grid(rmg, (-1., 1.), axes=(1, )), False)
